from abc import ABC, abstractmethod
from functools import cached_property
import numpy as np
from gpaw.response import timer
from scipy.spatial import Delaunay
from scipy.linalg.blas import zher

import gpaw.cgpaw as cgpaw
from gpaw.utilities.blas import rk, mmm
from gpaw.utilities.progressbar import ProgressBar
from gpaw.response.pw_parallelization import Blocks1D


class Integrand(ABC):
    @abstractmethod
    def matrix_element(self, point):
        ...

    @abstractmethod
    def eigenvalues(self, point):
        ...


def czher(alpha: float, x, A) -> None:
    """Hermetian rank-1 update of upper half of A.

    A += alpha * np.outer(x.conj(), x)

    """
    AT = A.T
    out = zher(alpha, x, 1, 1, 0, len(x), AT, 1)
    assert out is AT


class Integrator:
    def __init__(self, cell_cv, context, blockcomm, kncomm):
        """Baseclass for Brillouin zone integration and band summation.

        Simple class to calculate integrals over Brilloun zones
        and summation of bands.

        context: ResponseContext
        """
        self.vol = abs(np.linalg.det(cell_cv))
        self.context = context
        self.blockcomm = blockcomm
        self.kncomm = kncomm

    def mydomain(self, domain):
        from gpaw.response.pw_parallelization import Blocks1D
        # This function does the same as distribute_domain
        # but on a flat list and without all the fluff.
        #
        # In progress: Getting rid of distribute_domain(),
        blocks = Blocks1D(self.kncomm, len(domain))
        return [domain[i] for i in range(blocks.a, blocks.b)]

    def integrate(self, **kwargs):
        raise NotImplementedError


class PointIntegrator(Integrator):
    """Integrate brillouin zone using a broadening technique.

    The broadening technique consists of smearing out the
    delta functions appearing in many integrals by some factor
    eta. In this code we use Lorentzians."""

    def integrate(self, *, task, wd, domain, integrand, out_wxx):
        """Integrate a response function over bands and kpoints."""

        self.context.print('Integral kind:', task.kind)

        mydomain = self.mydomain(domain)

        prefactor = (2 * np.pi)**3 / self.vol / domain.nkpts
        out_wxx /= prefactor

        # Sum kpoints
        # Calculate integrations weight
        pb = ProgressBar(self.context.fd)
        for _, point in pb.enumerate(mydomain):
            n_MG = integrand.matrix_element(point)
            if n_MG is None:
                continue
            deps_M = integrand.eigenvalues(point)

            task.run(wd, n_MG, deps_M, out_wxx)

        # Sum over
        # Can this really be valid, if the original input out_wxx is nonzero?
        # This smells and should be investigated XXX
        # There could also be similar errors elsewhere... XXX
        self.kncomm.sum(out_wxx)

        if self.blockcomm.size == 1 and task.symmetrizable_unless_blocked:
            # Fill in upper/lower triangle also:
            nx = out_wxx.shape[1]
            il = np.tril_indices(nx, -1)
            iu = il[::-1]

            if isinstance(task, Hilbert):
                # XXX special hack since one of them wants the other
                # triangle.
                for out_xx in out_wxx:
                    out_xx[il] = out_xx[iu].conj()
            else:
                for out_xx in out_wxx:
                    out_xx[iu] = out_xx[il].conj()

        out_wxx *= prefactor


class IntegralTask(ABC):
    # Unique string for each kind of integral:
    kind = '(unset)'

    # Some integrals kinds like to calculate upper or lower half of the output
    # when nblocks==1.  In that case, this boolean signifies to the
    # integrator that the output array should be symmetrized.
    #
    # Actually: We don't gain anything much by doing this boolean
    # more systematically, since it's just Hermitian and Hilbert that need
    # it, and then one of the Tetrahedron types which is not compatible
    # anyway.  We should probably not do this.
    symmetrizable_unless_blocked = False

    @abstractmethod
    def run(self, wd, n_mG, deps_m, out_wxx):
        """Add contribution from one point to out_wxx."""


class GenericUpdate(IntegralTask):
    kind = 'response function'
    symmetrizable_unless_blocked = False

    def __init__(self, eta, blockcomm, eshift=None):
        self.eta = eta
        self.blockcomm = blockcomm
        self.eshift = eshift or 0.0

    # @timer('CHI_0 update')
    def run(self, wd, n_mG, deps_m, chi0_wGG):
        """Update chi."""

        deps_m += self.eshift * np.sign(deps_m)
        deps1_m = deps_m + 1j * self.eta
        deps2_m = deps_m - 1j * self.eta

        blocks1d = Blocks1D(self.blockcomm, chi0_wGG.shape[2])

        for omega, chi0_GG in zip(wd.omega_w, chi0_wGG):
            x_m = (1 / (omega + deps1_m) - 1 / (omega - deps2_m))
            if blocks1d.blockcomm.size > 1:
                nx_mG = n_mG[:, blocks1d.myslice] * x_m[:, np.newaxis]
            else:
                nx_mG = n_mG * x_m[:, np.newaxis]

            mmm(1.0, np.ascontiguousarray(nx_mG.T), 'N', n_mG.conj(), 'N',
                1.0, chi0_GG)


class Hermitian(IntegralTask):
    kind = 'hermitian response function'
    symmetrizable_unless_blocked = True

    def __init__(self, blockcomm, eshift=None):
        self.blockcomm = blockcomm
        self.eshift = eshift or 0.0

    # @timer('CHI_0 hermetian update')
    def run(self, wd, n_mG, deps_m, chi0_wGG):
        """If eta=0 use hermitian update."""
        deps_m += self.eshift * np.sign(deps_m)

        blocks1d = Blocks1D(self.blockcomm, chi0_wGG.shape[2])

        for w, omega in enumerate(wd.omega_w):
            if blocks1d.blockcomm.size == 1:
                x_m = np.abs(2 * deps_m / (omega.imag**2 + deps_m**2))**0.5
                nx_mG = n_mG.conj() * x_m[:, np.newaxis]
                rk(-1.0, nx_mG, 1.0, chi0_wGG[w], 'n')
            else:
                x_m = np.abs(2 * deps_m / (omega.imag**2 + deps_m**2))
                mynx_mG = n_mG[:, blocks1d.myslice] * x_m[:, np.newaxis]
                mmm(-1.0, mynx_mG, 'T', n_mG.conj(), 'N', 1.0, chi0_wGG[w])


class Hilbert(IntegralTask):
    kind = 'spectral function'
    symmetrizable_unless_blocked = True

    def __init__(self, blockcomm, eshift=None):
        self.blockcomm = blockcomm
        self.eshift = eshift or 0.0

    # @timer('CHI_0 spectral function update (new)')
    def run(self, wd, n_mG, deps_m, chi0_wGG):
        """Update spectral function.

        Updates spectral function A_wGG and saves it to chi0_wGG for
        later hilbert-transform."""

        deps_m += self.eshift * np.sign(deps_m)
        o_m = abs(deps_m)
        w_m = wd.get_floor_index(o_m)

        blocks1d = Blocks1D(self.blockcomm, chi0_wGG.shape[2])

        # Sort frequencies
        argsw_m = np.argsort(w_m)
        sortedo_m = o_m[argsw_m]
        sortedw_m = w_m[argsw_m]
        sortedn_mG = n_mG[argsw_m]

        index = 0
        while 1:
            if index == len(sortedw_m):
                break

            w = sortedw_m[index]
            startindex = index
            while 1:
                index += 1
                if index == len(sortedw_m):
                    break
                if w != sortedw_m[index]:
                    break

            endindex = index

            # Here, we have same frequency range w, for set of
            # electron-hole excitations from startindex to endindex.
            o1 = wd.omega_w[w]
            o2 = wd.omega_w[w + 1]
            p = np.abs(1 / (o2 - o1)**2)
            p1_m = np.array(p * (o2 - sortedo_m[startindex:endindex]))
            p2_m = np.array(p * (sortedo_m[startindex:endindex] - o1))

            if blocks1d.blockcomm.size > 1 and w + 1 < wd.wmax:
                x_mG = sortedn_mG[startindex:endindex, blocks1d.myslice]
                mmm(1.0,
                    np.concatenate((p1_m[:, None] * x_mG,
                                    p2_m[:, None] * x_mG),
                                   axis=1).T.copy(),
                    'N',
                    sortedn_mG[startindex:endindex].T.copy(),
                    'C',
                    1.0,
                    chi0_wGG[w:w + 2].reshape((2 * blocks1d.nlocal,
                                               blocks1d.N)))

            if blocks1d.blockcomm.size <= 1 and w + 1 < wd.wmax:
                x_mG = sortedn_mG[startindex:endindex]
                l_Gm = (p1_m[:, None] * x_mG).T.copy()
                r_Gm = x_mG.T.copy()
                mmm(1.0, r_Gm, 'N', l_Gm, 'C', 1.0, chi0_wGG[w])
                l_Gm = (p2_m[:, None] * x_mG).T.copy()
                mmm(1.0, r_Gm, 'N', l_Gm, 'C', 1.0, chi0_wGG[w + 1])


class Intraband(IntegralTask):
    kind = 'intraband'
    symmetrizable_unless_blocked = False

    # @timer('CHI_0 intraband update')
    def run(self, wd, vel_mv, deps_M, chi0_wvv):
        """Add intraband contributions"""
        # Intraband is a little bit special, we use neither wd nor deps_M

        for vel_v in vel_mv:
            x_vv = np.outer(vel_v, vel_v)
            chi0_wvv[0] += x_vv


class OpticalLimit(IntegralTask):
    kind = 'response function wings'
    symmetrizable_unless_blocked = False

    def __init__(self, eta):
        self.eta = eta

    # @timer('CHI_0 optical limit update')
    def run(self, wd, n_mG, deps_m, chi0_wxvG):
        """Optical limit update of chi."""
        deps1_m = deps_m + 1j * self.eta
        deps2_m = deps_m - 1j * self.eta

        for w, omega in enumerate(wd.omega_w):
            x_m = (1 / (omega + deps1_m) - 1 / (omega - deps2_m))
            chi0_wxvG[w, 0] += np.dot(x_m * n_mG[:, :3].T, n_mG.conj())
            chi0_wxvG[w, 1] += np.dot(x_m * n_mG[:, :3].T.conj(), n_mG)


class HermitianOpticalLimit(IntegralTask):
    kind = 'hermitian response function wings'
    symmetrizable_unless_blocked = False

    # @timer('CHI_0 hermitian optical limit update')
    def run(self, wd, n_mG, deps_m, chi0_wxvG):
        """Optical limit update of hermitian chi."""
        for w, omega in enumerate(wd.omega_w):
            x_m = - np.abs(2 * deps_m / (omega.imag**2 + deps_m**2))
            chi0_wxvG[w, 0] += np.dot(x_m * n_mG[:, :3].T, n_mG.conj())
            chi0_wxvG[w, 1] += np.dot(x_m * n_mG[:, :3].T.conj(), n_mG)


class HilbertOpticalLimit(IntegralTask):
    kind = 'spectral function wings'
    symmetrizable_unless_blocked = False

    # @timer('CHI_0 optical limit hilbert-update')
    def run(self, wd, n_mG, deps_m, chi0_wxvG):
        """Optical limit update of chi-head and -wings."""

        for deps, n_G in zip(deps_m, n_mG):
            o = abs(deps)
            w = wd.get_floor_index(o)
            if w + 1 >= wd.wmax:
                continue
            o1, o2 = wd.omega_w[w:w + 2]
            if o > o2:
                continue
            else:
                assert o1 <= o <= o2, (o1, o, o2)

            p = 1 / (o2 - o1)**2
            p1 = p * (o2 - o)
            p2 = p * (o - o1)
            x_vG = np.outer(n_G[:3], n_G.conj())
            chi0_wxvG[w, 0, :, :] += p1 * x_vG
            chi0_wxvG[w + 1, 0, :, :] += p2 * x_vG
            chi0_wxvG[w, 1, :, :] += p1 * x_vG.conj()
            chi0_wxvG[w + 1, 1, :, :] += p2 * x_vG.conj()


class Point:
    def __init__(self, kpt_c, K, spin):
        self.kpt_c = kpt_c
        self.K = K
        self.spin = spin


class Domain:
    def __init__(self, kpts_kc, spins):
        self.kpts_kc = kpts_kc
        self.spins = spins

    @property
    def nkpts(self):
        return len(self.kpts_kc)

    @property
    def nspins(self):
        return len(self.spins)

    def __len__(self):
        return self.nkpts * self.nspins

    def __getitem__(self, num) -> Point:
        K = num // self.nspins
        return Point(self.kpts_kc[K], K,
                     self.spins[num % self.nspins])

    def tesselation(self):
        tesselation = KPointTesselation(self.kpts_kc)
        tesselated_domains = Domain(tesselation.bzk_kc, self.spins)
        return tesselation, tesselated_domains


class KPointTesselation:
    def __init__(self, kpts):
        self._td = Delaunay(kpts)

    @property
    def bzk_kc(self):
        return self._td.points

    @cached_property
    def simplex_volumes(self):
        volumes_s = np.zeros(self._td.nsimplex, float)
        for s in range(self._td.nsimplex):
            K_k = self._td.simplices[s]
            k_kc = self._td.points[K_k]
            volume = np.abs(np.linalg.det(k_kc[1:] - k_kc[0])) / 6.
            volumes_s[s] = volume
        return volumes_s

    def tetrahedron_weight(self, K, deps_k, omega_w):
        simplices_s = self.pts_k[K]
        W_w = np.zeros(len(omega_w), float)
        vol_s = self.simplex_volumes[simplices_s]
        cgpaw.tetrahedron_weight(
            deps_k, self._td.simplices, K, simplices_s, W_w, omega_w, vol_s)
        return W_w

    @cached_property
    def pts_k(self):
        pts_k = [[] for n in range(self.nkpts)]
        for s, K_k in enumerate(self._td.simplices):
            A_kv = np.append(self._td.points[K_k],
                             np.ones(4)[:, np.newaxis], axis=1)

            D_kv = np.append((A_kv[:, :-1]**2).sum(1)[:, np.newaxis],
                             A_kv, axis=1)
            a = np.linalg.det(D_kv[:, np.arange(5) != 0])

            if np.abs(a) < 1e-10:
                continue

            for K in K_k:
                pts_k[K].append(s)

        return [np.array(pts_k[k], int) for k in range(self.nkpts)]

    @property
    def nkpts(self):
        return self._td.npoints

    @cached_property
    def neighbours_k(self):
        return [np.unique(self._td.simplices[self.pts_k[k]])
                for k in range(self.nkpts)]


class TetrahedronIntegrator(Integrator):
    """Integrate brillouin zone using tetrahedron integration.

    Tetrahedron integration uses linear interpolation of
    the eigenenergies and of the matrix elements
    between the vertices of the tetrahedron."""

    @timer('Spectral function integration')
    def integrate(self, *, domain, integrand, wd, out_wxx, task):
        """Integrate response function.

        Assume that the integral has the
        form of a response function. For the linear tetrahedron
        method it is possible calculate frequency dependent weights
        and do a point summation using these weights."""

        tesselation, alldomains = domain.tesselation()
        mydomain = self.mydomain(alldomains)

        with self.context.timer('eigenvalues'):
            deps_tMk = None  # t for term

            for point in alldomains:
                deps_M = -integrand.eigenvalues(point)
                if deps_tMk is None:
                    deps_tMk = np.zeros([alldomains.nspins, *deps_M.shape,
                                         tesselation.nkpts], float)
                deps_tMk[point.spin, :, point.K] = deps_M

        # Calculate integrations weight
        pb = ProgressBar(self.context.fd)
        for _, point in pb.enumerate(mydomain):
            deps_Mk = deps_tMk[point.spin]
            teteps_Mk = deps_Mk[:, tesselation.neighbours_k[point.K]]
            n_MG = integrand.matrix_element(point)

            # Generate frequency weights
            i0_M, i1_M = wd.get_index_range(teteps_Mk.min(1), teteps_Mk.max(1))
            with self.context.timer('tetrahedron weight'):
                W_Mw = []
                for deps_k, i0, i1 in zip(deps_Mk, i0_M, i1_M):
                    W_w = tesselation.tetrahedron_weight(
                        point.K, deps_k, wd.omega_w[i0:i1])
                    W_Mw.append(W_w)

            task.run(n_MG, deps_Mk, W_Mw, i0_M, i1_M, out_wxx)

        self.kncomm.sum(out_wxx)

        if self.blockcomm.size == 1 and task.symmetrizable_unless_blocked:
            # Fill in upper/lower triangle also:
            nx = out_wxx.shape[1]
            il = np.tril_indices(nx, -1)
            iu = il[::-1]
            for out_xx in out_wxx:
                out_xx[il] = out_xx[iu].conj()


class HilbertTetrahedron:
    kind = 'spectral function'
    symmetrizable_unless_blocked = True

    def __init__(self, blockcomm):
        self.blockcomm = blockcomm

    def run(self, n_MG, deps_Mk, W_Mw, i0_M, i1_M, out_wxx):
        """Update output array with dissipative part."""
        blocks1d = Blocks1D(self.blockcomm, out_wxx.shape[2])

        for n_G, deps_k, W_w, i0, i1 in zip(n_MG, deps_Mk, W_Mw,
                                            i0_M, i1_M):
            if i0 == i1:
                continue

            for iw, weight in enumerate(W_w):
                if blocks1d.blockcomm.size > 1:
                    myn_G = n_G[blocks1d.myslice].reshape((-1, 1))
                    # gemm(weight, n_G.reshape((-1, 1)), myn_G,
                    #      1.0, out_wxx[i0 + iw], 'c')
                    mmm(weight, myn_G, 'N', n_G.reshape((-1, 1)), 'C',
                        1.0, out_wxx[i0 + iw])
                else:
                    czher(weight, n_G.conj(), out_wxx[i0 + iw])


class HilbertOpticalLimitTetrahedron:
    kind = 'spectral function wings'
    symmetrizable_unless_blocked = False

    def run(self, n_MG, deps_Mk, W_Mw, i0_M, i1_M, out_wxvG):
        """Update optical limit output array with dissipative part of the head
        and wings."""
        for n_G, deps_k, W_w, i0, i1 in zip(n_MG, deps_Mk, W_Mw,
                                            i0_M, i1_M):
            if i0 == i1:
                continue

            for iw, weight in enumerate(W_w):
                x_vG = np.outer(n_G[:3], n_G.conj())
                out_wxvG[i0 + iw, 0, :, :] += weight * x_vG
                out_wxvG[i0 + iw, 1, :, :] += weight * x_vG.conj()



import pytest
from gpaw import GPAW, restart


@pytest.mark.mom
def test_mom_lcao_smearing(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['co_mom'])
    E_gs = calc.get_potential_energy()

    f_sn = []
    for spin in range(calc.get_number_of_spins()):
        f_n = calc.get_occupation_numbers(spin=spin)
        f_sn.append(f_n)

    ne0_gs = f_sn[0].sum()
    f_sn[0][3] -= 1.
    f_sn[0][5] += 1.

    # Test both MOM and fixed occupations with Gaussian smearing
    for i in [True, False]:
        atoms, calc = restart(gpw_files['co_mom'], txt='-')

        # Excited-state calculation with Gaussian
        # smearing of the occupation numbers
        calc.set(occupations={'name': 'mom', 'numbers': f_sn,
                              'width': 0.01, 'use_fixed_occupations': i})
        E_es = atoms.get_potential_energy()

        f_n0 = calc.get_occupation_numbers(spin=0)
        ne0_es = f_n0.sum()

        dE = E_es - E_gs
        dne0 = ne0_es - ne0_gs
        print(dE)
        assert dE == pytest.approx(9.8445603513, abs=0.01)
        assert dne0 == pytest.approx(0.0, abs=1e-16)



"""
This file contains several routines to do the non linear
interpolation of poles and residues of respose fuctions
corresponding to the Multipole Approximation (MPA)
developed in the Ref. [1].

The implemented solver is the one based on Pade-Thiele
formula (See App. A of Ref. [1]).

[1] DA. Leon et al, PRB 104, 115157 (2021)
"""
from __future__ import annotations
from typing import Tuple, no_type_check
from gpaw.typing import Array1D, Array2D, Array3D
import numpy as np
from numpy.linalg import eigvals


def fit_residue(
    npr_GG: Array2D, omega_w: Array1D, X_wGG: Array3D, E_pGG: Array3D
) -> Array3D:
    npols = len(E_pGG)
    nw = len(omega_w)
    A_GGwp = np.zeros((*E_pGG.shape[1:], nw, npols), dtype=np.complex128)
    b_GGw = np.zeros((*E_pGG.shape[1:], nw), dtype=np.complex128)
    for w in range(nw):
        A_GGwp[:, :, w, :] = (
            2 * E_pGG / (omega_w[w]**2 - E_pGG**2)).transpose((1, 2, 0))
        b_GGw[:, :, w] = X_wGG[w, :, :]

    for p in range(npols):
        for w in range(A_GGwp.shape[2]):
            A_GGwp[:, :, w, p][p >= npr_GG] = 0.0

    temp_GGp = np.einsum('GHwp,GHw->GHp',
                         A_GGwp.conj(), b_GGw)
    XTX_GGpp = np.einsum('GHwp,GHwo->GHpo',
                         A_GGwp.conj(), A_GGwp)

    if XTX_GGpp.shape[2] == 1:
        # 1D matrix, invert the number
        XTX_GGpp = 1 / XTX_GGpp
        R_GGp = np.einsum('GHpo,GHo->GHp',
                          XTX_GGpp, temp_GGp)
    else:
        try:
            R_GGp = np.linalg.solve(XTX_GGpp, temp_GGp)
        except np.linalg.LinAlgError:
            XTX_GGpp = np.linalg.pinv(XTX_GGpp)
            R_GGp = np.einsum('GHpo,GHo->GHp',
                              XTX_GGpp, temp_GGp)

    return R_GGp.transpose((2, 0, 1))


class Solver:
    """
    X(w) is approximated as a sum of poles
    Form of one pole: 2*E*R/(w**2-E**2)
    The input are two w and X(w) for each pole
    The output are E and R coefficients
    """
    def __init__(self, omega_w: Array1D, threshold=1e-5, epsilon=1e-8):
        """
        Parameters
        ----------
        omega_w : Array of complex frequencies set in
                  mpa sampling. The length corresponds
                  to twice the number of poles
        threshold: threshold for small and too close poles
        epsilon: precision for positive zero imaginary part
        """
        assert len(omega_w) % 2 == 0
        self.omega_w = omega_w
        self.npoles = len(omega_w) // 2
        self.threshold = threshold
        self.epsilon = epsilon

    def solve(self, X_wGG):
        """
        X_wGG is any response function evaluated at omega_w
        it returns a tuple of poles and residues (E_pGG, R_pGG)
        where p is the pole index
        """
        raise NotImplementedError


class SinglePoleSolver(Solver):
    def __init__(self, omega_w: Array1D):
        Solver.__init__(self, omega_w=omega_w)

    def solve(self, X_wGG: Array3D) -> Tuple[Array2D, Array2D]:
        """
        This interpolates X_wGG using a single pole (E_GG, R_GG)
        """
        assert len(X_wGG) == 2

        omega_w = self.omega_w
        E_GG = ((X_wGG[0, :, :] * omega_w[0]**2 -
                 X_wGG[1, :, :] * omega_w[1]**2) /
                (X_wGG[0, :, :] - X_wGG[1, :, :])
                )  # analytical solution

        def branch_sqrt_inplace(E_GG: Array2D):
            E_GG.real = np.abs(E_GG.real)  # physical pole
            E_GG.imag = -np.abs(E_GG.imag)  # correct time ordering
            E_GG[:] = np.emath.sqrt(E_GG)

        branch_sqrt_inplace(E_GG)
        mask = E_GG < self.threshold  # null pole
        E_GG[mask] = self.threshold - 1j * self.epsilon

        R_GG = fit_residue(
            npr_GG=np.zeros_like(E_GG) + 1,
            omega_w=omega_w,
            X_wGG=X_wGG,
            E_pGG=E_GG.reshape((1, *E_GG.shape)))[0, :, :]

        return E_GG, R_GG


class MultipoleSolver(Solver):
    def __init__(self, omega_w: Array1D):
        Solver.__init__(self, omega_w=omega_w)

    def solve(self, X_wGG: Array3D) -> Tuple[Array3D, Array3D]:
        """
        This interpolates X_wGG using a sveral poles (E_pGG, R_pGG)
        """
        assert len(X_wGG) == 2 * self.npoles

        # First the poles are obtained (non linear part of the problem)
        E_GGp, npr_GG = pade_solve(X_wGG, self.omega_w**2)
        E_pGG = E_GGp.transpose((2, 0, 1))
        # The residues are obtained in a linear least square problem with
        # complex variables
        R_pGG = fit_residue(npr_GG, self.omega_w, X_wGG, E_pGG)
        return E_pGG, R_pGG


def RESolver(omega_w: Array1D):
    assert len(omega_w) % 2 == 0
    npoles = len(omega_w) / 2
    assert npoles > 0
    if npoles == 1:
        return SinglePoleSolver(omega_w)
    else:
        return MultipoleSolver(omega_w)


def mpa_cond_vectorized(
    npols: int, z_w: Array1D, E_GGp: Array3D, pole_resolution: float = 1e-5
) -> Tuple[Array3D, Array2D]:
    wmax = np.max(np.real(np.emath.sqrt(z_w))) * 1.5

    E_GGp = np.emath.sqrt(E_GGp)
    args = np.abs(E_GGp.real), np.abs(E_GGp.imag)
    E_GGp = np.maximum(*args) - 1j * np.minimum(*args)
    E_GGp.sort(axis=2)  # Sort according to real part

    for i in range(npols):
        out_poles_GG = E_GGp[:, :, i].real > wmax
        E_GGp[out_poles_GG, i] = 2 * wmax - 0.01j
        for j in range(i + 1, npols):
            diff = E_GGp[:, :, j].real - E_GGp[:, :, i].real
            equal_poles_GG = diff < pole_resolution
            if np.sum(equal_poles_GG.ravel()):
                break

            # if the poles are to close, move them to the end, set value
            # to sort to the end of array (e.g., 2x wmax)
            E_GGp[:, :, j] = np.where(
                equal_poles_GG,
                (E_GGp[:, :, j].real + E_GGp[:, :, i].real) / 2
                + 1j * np.maximum(E_GGp[:, :, i].imag, E_GGp[:, :, j].imag),
                E_GGp[:, :, j],
            )
            E_GGp[equal_poles_GG, i] = 2 * wmax - 0.01j

    E_GGp.sort(axis=2)  # Sort according to real part

    npr_GG = np.sum(E_GGp.real < wmax, axis=2)
    return E_GGp, npr_GG


@no_type_check
def pade_solve(X_wGG: Array3D, z_w: Array1D) -> Tuple[Array3D, Array2D]:
    nw, nG, _ = X_wGG.shape
    npols = nw // 2
    nm = npols + 1
    b_GGm = np.zeros((nG, nG, nm), dtype=np.complex128)
    b_GGm[..., 0] = 1.0
    bm1_GGm = b_GGm
    c_GGw = X_wGG.transpose((1, 2, 0)).copy()

    for i in range(1, 2 * npols):
        cm1_GGw = np.copy(c_GGw)
        bm2_GGm = np.copy(bm1_GGm)
        bm1_GGm = np.copy(b_GGm)

        current_z = z_w[i - 1]

        c_GGw[..., i:] = (
            (cm1_GGw[..., i - 1, np.newaxis] - cm1_GGw[..., i:]) /
            ((z_w[i:] - current_z) * cm1_GGw[..., i:])
        )

        b_GGm -= current_z * c_GGw[..., i, np.newaxis] * bm2_GGm
        bm2_GGm[..., 1:] = c_GGw[..., i, np.newaxis] * bm2_GGm[..., :-1]
        b_GGm[..., 1:] += bm2_GGm[..., 1:]

    companion_GGpp = np.zeros((nG, nG, npols, npols),
                              dtype=np.complex128)

    # Create a poly companion matrix in vectorized form
    # Equal to following serial code
    # for i in range(nG):
    #     for j in range(nG):
    #         companion_GGpp[i, j] = poly.polycompanion(b_GGm[i, j])
    b_GGm /= b_GGm[:, :, -1][..., None]
    companion_GGpp.reshape((nG, nG, -1))[:, :, npols::npols + 1] = 1
    companion_GGpp[:, :, :, -1] = -b_GGm[:, :, :npols]

    E_GGp = eigvals(companion_GGpp)
    E_GGp, npr_GG = mpa_cond_vectorized(npols=npols, z_w=z_w, E_GGp=E_GGp)
    return E_GGp, npr_GG



from __future__ import annotations
from math import pi

import numpy as np
from gpaw.core import PWArray, PWDesc, UGDesc
from gpaw.core.arrays import DistributedArrays as XArray
from gpaw.core.atom_arrays import AtomArrays
from gpaw.hybrids.paw import pawexxvv
from gpaw.hybrids.wstc import WignerSeitzTruncatedCoulomb
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.pw.hamiltonian import PWHamiltonian
from gpaw.typing import Array1D
from gpaw.utilities import unpack_hermitian


def coulomb(pw: PWDesc, grid: UGDesc, omega: float):
    if omega == 0.0:
        wstc = WignerSeitzTruncatedCoulomb(
            pw.cell_cv, np.array([1, 1, 1]))
        return wstc.get_potential_new(pw, grid)

    v_G = pw.empty()
    G2_G = pw.ekin_G * 2
    v_G.data[:] = 4 * pi * (1 - np.exp(-G2_G / (4 * omega**2)))
    if pw.ng1 == 0:
        v_G.data[1:] /= G2_G[1:]
        v_G.data[0] = pi / omega**2
    else:
        v_G.data /= G2_G

    return v_G


class PWHybridHamiltonian(PWHamiltonian):
    def __init__(self,
                 grid: UGDesc,
                 pw: PWDesc,
                 xc,
                 setups,
                 fracpos_ac,
                 atomdist):
        super().__init__(grid, pw)
        self.grid = grid
        self.pw = pw
        self.exx_fraction = xc.exx_fraction
        self.exx_omega = xc.exx_omega

        self.exx_cc = sum(setup.ExxC for setup in setups) * self.exx_fraction
        self.VC_aii = [unpack_hermitian(setup.X_p * self.exx_fraction)
                       for setup in setups]
        self.delta_aiiL = [setup.Delta_iiL for setup in setups]
        self.C_app = [setup.M_pp * self.exx_fraction for setup in setups]

        self.v_G = coulomb(pw, grid, self.exx_omega)
        self.v_G.data *= self.exx_fraction
        self.ghat_aLG = setups.create_compensation_charges(
            pw, fracpos_ac, atomdist)
        self.plan = grid.fft_plans()

    def apply_orbital_dependent(self,
                                ibzwfs: IBZWaveFunctions,
                                D_asii,
                                psit2_nG: XArray,
                                spin: int,
                                Htpsit2_nG: XArray) -> None:
        assert isinstance(psit2_nG, PWArray)
        assert isinstance(Htpsit2_nG, PWArray)
        wfs = ibzwfs.wfs_qs[0][spin]
        D_aii = D_asii[:, spin]
        if ibzwfs.nspins == 1:
            D_aii.data *= 0.5
        f1_n = wfs.myocc_n
        psit1_nG = wfs.psit_nX
        P1_ani = wfs.P_ani
        pt_aiG = wfs.pt_aiX

        same = psit1_nG is psit2_nG

        if same:
            P2_ani = P1_ani
        else:
            P2_ani = pt_aiG.integrate(psit2_nG)

        evv, evc, ekin = self.calculate(D_aii, same, pt_aiG,
                                        psit1_nG, P1_ani, f1_n,
                                        psit2_nG, P2_ani,
                                        Htpsit2_nG)
        if same:
            for name, e in [('exx_vv', evv),
                            ('exx_vc', evc),
                            ('exx_kinetic', ekin)]:
                e *= ibzwfs.spin_degeneracy
                if spin == 0:
                    ibzwfs.energies[name] = e
                else:
                    ibzwfs.energies[name] += e
            ibzwfs.energies['exx_cc'] = self.exx_cc

    def calculate(self,
                  D_aii,
                  same,
                  pt_aiG,
                  psit1_nG: PWArray,
                  P1_ani: AtomArrays,
                  f1_n: Array1D,
                  psit2_nG: PWArray,
                  P2_ani: AtomArrays,
                  Htpsit2_nG: PWArray):

        psit1_R = self.grid.empty()
        psit2_R = self.grid.empty()
        rhot_R = self.grid.empty()
        rhot_G = self.pw.empty()
        vrhot_G = self.pw.empty()

        evv = 0.0
        evc = 0.0
        ekin = 0.0
        B_ani = {}
        for a, D_ii in D_aii.items():
            VV_ii = pawexxvv(self.C_app[a], D_ii)
            VC_ii = self.VC_aii[a]
            B_ni = P2_ani[a] @ (-VC_ii - 2 * VV_ii)
            B_ani[a] = B_ni
            if same:
                ec = (D_ii * VC_ii).sum()
                ev = (D_ii * VV_ii).sum()
                ekin += ec + 2 * ev
                evv -= ev
                evc -= ec

        for n2, (psit2_G, out_G) in enumerate(zip(psit2_nG, Htpsit2_nG)):
            psit2_G.ifft(out=psit2_R, plan=self.plan)

            for n1, (psit1_G, f1) in enumerate(zip(psit1_nG, f1_n)):
                psit1_G.ifft(out=psit1_R, plan=self.plan)
                rhot_R.data[:] = psit1_R.data * psit2_R.data
                rhot_R.fft(out=rhot_G, plan=self.plan)
                Q_aL = {a: np.einsum('i, ijL, j -> L',
                                     P1_ani[a][n1], delta_iiL, P2_ani[a][n2])
                        for a, delta_iiL in enumerate(self.delta_aiiL)}
                self.ghat_aLG.add_to(rhot_G, Q_aL)
                vrhot_G.data[:] = rhot_G.data * self.v_G.data
                if same:
                    e = f1 * f1_n[n2] * rhot_G.integrate(vrhot_G)
                    evv -= 0.5 * e
                    ekin += e
                vrhot_G.ifft(out=rhot_R, plan=self.plan)
                rhot_R.data *= psit1_R.data
                rhot_R.fft(out=rhot_G, plan=self.plan)
                out_G.data -= rhot_G.data * f1

                A_aL = self.ghat_aLG.integrate(vrhot_G)
                for a, A_L in A_aL.items():
                    B_ani[a][n2] -= np.einsum(
                        'L, ijL, j -> i',
                        f1 * A_L, self.delta_aiiL[a], P1_ani[a][n1])

        pt_aiG.add_to(Htpsit2_nG, B_ani)

        return evv, evc, ekin



import pytest

import numpy as np

from gpaw import GPAW
from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator, SelfEnhancementCalculator
from gpaw.response.dyson import DysonSolver
from gpaw.response.goldstone import NewFMGoldstoneScaling
from gpaw.response.susceptibility import (spectral_decomposition,
                                          read_eigenmode_lineshapes)

from gpaw.test import findpeak
from gpaw.test.gpwfile import response_band_cutoff


@pytest.mark.kspair
@pytest.mark.response
def test_response_cobalt_sf_gsspawALDA(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    q_qc = [[0.0, 0.0, 0.0], [1. / 4., 0.0, 0.0]]  # Two q-points along G-M
    frq_w = np.linspace(-0.5, 2.0, 101)
    eta = 0.2

    rshelmax = 0
    ecut = 150
    pos_eigs = 5
    nmodes = 2  # majority modes
    nblocks = 'max'

    # ---------- Script ---------- #

    # Read ground state data
    context = ResponseContext(txt='cobalt_susceptibility.txt')
    calc = GPAW(gpw_files['co_pw'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw']
    gs = ResponseGroundStateAdapter(calc)

    # Set up response calculators
    calc_args = (gs,)
    calc_kwargs = dict(context=context,
                       nbands=nbands,
                       ecut=ecut,
                       gammacentered=True,
                       bandsummation='pairwise',
                       nblocks=nblocks)
    chiks_calc = ChiKSCalculator(*calc_args, **calc_kwargs)
    xi_calc = SelfEnhancementCalculator(*calc_args,
                                        rshelmax=rshelmax,
                                        **calc_kwargs)
    hxc_scaling = NewFMGoldstoneScaling.from_xi_calculator(xi_calc)
    dyson_solver = DysonSolver(context)

    for q, q_c in enumerate(q_qc):
        # Calculate χ_KS^+-(q,z) and Ξ^++(q,z)
        zd = ComplexFrequencyDescriptor.from_array(frq_w + 1j * eta)
        chiks = chiks_calc.calculate('+-', q_c, zd)
        xi = xi_calc.calculate('+-', q_c, zd)

        # Distribute frequencies and invert dyson equation
        chiks = chiks.copy_with_global_frequency_distribution()
        xi = xi.copy_with_global_frequency_distribution()
        chi = dyson_solver(chiks, xi)
        # Test ability to apply Goldstone scaling when inverting the dyson eq.
        scaled_chi = dyson_solver(chiks, xi, hxc_scaling=hxc_scaling)
        # Simulate a "restart" of the GoldstoneScaling
        hxc_scaling = NewFMGoldstoneScaling(lambd=hxc_scaling.lambd)

        # Calculate majority spectral function
        Amaj, _ = spectral_decomposition(chi, pos_eigs=pos_eigs)
        Amaj.write_eigenmode_lineshapes(
            f'cobalt_Amaj_q{q}.csv', nmodes=nmodes)
        sAmaj, _ = spectral_decomposition(scaled_chi, pos_eigs=pos_eigs)
        sAmaj.write_eigenmode_lineshapes(
            f'cobalt_sAmaj_q{q}.csv', nmodes=nmodes)

        # Store Re ξ^++(q=0,ω), to test the self-enhancement after scaling
        if q == 0:
            chiks_mW, xi_mW = get_mode_projections(
                chiks, xi, sAmaj, lambd=hxc_scaling.lambd, nmodes=nmodes)
            xi0_w = xi_mW[0].real

        # plot_enhancement(chiks, xi, Amaj, sAmaj,
        #                  lambd=hxc_scaling.lambd, nmodes=nmodes)

    context.write_timer()

    # Compare scaling coefficient to reference
    assert hxc_scaling.lambd == pytest.approx(1.0541, abs=0.001)
    # Test that Re ξ^++(q=0,ω) ≾ 1 at ω=0
    w0 = np.argmin(np.abs(frq_w))
    assert xi0_w[w0] == pytest.approx(0.987, abs=0.01)

    # Compare magnon peaks to reference data
    refs_mqa = [
        # Acoustic
        [
            # q_Γ
            [
                # (wpeak, Apeak)
                (0.085, 7.895),  # unscaled
                (-0.002, 7.980),  # scaled
            ],
            # q_M / 2
            [
                # (wpeak, Apeak)
                (0.320, 5.828),  # unscaled
                (0.245, 6.215),  # scaled
            ],
        ],
        # Optical
        [
            # q_Γ
            [
                # (wpeak, Apeak)
                (0.904, 3.493),  # unscaled
                (0.860, 3.395),  # scaled
            ],
            # q_M / 2
            [
                # (wpeak, Apeak)
                (0.857, 2.988),  # unscaled
                (0.721, 3.163),  # scaled
            ],
        ],
    ]
    for a, Astr in enumerate(['Amaj', 'sAmaj']):
        for q in range(len(q_qc)):
            w_w, a_wm = read_eigenmode_lineshapes(f'cobalt_{Astr}_q{q}.csv')
            for m in range(nmodes):
                wpeak, Apeak = findpeak(w_w, a_wm[:, m])
                refw, refA = refs_mqa[m][q][a]
                print(m, q, a, wpeak, Apeak)
                assert wpeak == pytest.approx(refw, abs=0.01)  # eV
                assert Apeak == pytest.approx(refA, abs=0.05)  # a.u.


def get_mode_projections(chiks, xi, Amaj, *, lambd, nmodes):
    """Project χ_KS^+-(q,z) and Ξ^++(q,z) onto the magnon mode vectors."""
    wm = Amaj.get_eigenmode_frequency(nmodes=nmodes)
    v_Gm = Amaj.get_eigenvectors_at_frequency(wm, nmodes=nmodes)
    chiks_wm = np.zeros((chiks.blocks1d.nlocal, nmodes), dtype=complex)
    xi_wm = np.zeros((xi.blocks1d.nlocal, nmodes), dtype=complex)
    for m, v_G in enumerate(v_Gm.T):
        chiks_wm[:, m] = np.conj(v_G) @ chiks.array @ v_G  # chiks_wGG
        xi_wm[:, m] = np.conj(v_G) @ xi.array @ v_G  # xi_wGG
    chiks_mW = chiks.blocks1d.all_gather(chiks_wm * lambd).T
    xi_mW = xi.blocks1d.all_gather(xi_wm * lambd).T
    return chiks_mW, xi_mW


def plot_enhancement(chiks, xi, Amaj0, sAmaj, *, lambd, nmodes):
    import matplotlib.pyplot as plt
    from gpaw.mpi import world
    from ase.units import Ha

    for Amaj, _lambd in zip([Amaj0, sAmaj], [1., lambd]):
        a_mW = Amaj.get_eigenmode_lineshapes(nmodes=nmodes).T
        chiks_mW, xi_mW = get_mode_projections(
            chiks, xi, Amaj, lambd=_lambd, nmodes=nmodes)
        for m in range(nmodes):
            plt.subplot(1, nmodes, m + 1)
            plt.plot(chiks.zd.omega_w * Ha, -chiks_mW[m].imag / np.pi)
            plt.plot(xi.zd.omega_w * Ha, xi_mW[m].real)
            plt.axhline(1., c='0.5')
            plt.plot(Amaj.omega_w, a_mW[m])

    if world.rank == 0:
        plt.show()
    world.barrier()



import pytest

import numpy as np

from gpaw import GPAW
import gpaw.mpi as mpi
from gpaw.response.df import DielectricFunction, read_response_function

from gpaw.test import findpeak


@pytest.mark.dielectricfunction
@pytest.mark.response
def test_mos2_polarizability(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['mos2_pw'], communicator=mpi.serial_comm)

    # Calculate the 2D polarizability within the RPA and ALDA
    df = DielectricFunction(calc, truncation='2D',
                            frequencies={'type': 'nonlinear', 'domega0': 0.05},
                            integrationmode='tetrahedron integration',
                            nblocks='max')
    df.get_polarizability(xc='RPA', filename='rpa_pol.csv')
    mpi.world.barrier()  # give rank 0 some time to write the files

    # Test against reference values
    refs = [  # rpa
        # w0r, w0i, wr, wi
        [1.875, 2.745, 1.949, 2.856],
        # I0r, I0i, Ir, Ii
        [10.602, 11.145, 9.886, 10.866],
    ]
    omega_w, alpha0_w, alpha_w = read_response_function('rpa_pol.csv')
    # plot_pol(omega_w, alpha0_w, alpha_w)
    w0r, I0r, w0i, I0i = identify_maxima(omega_w, alpha0_w)
    wr, Ir, wi, Ii = identify_maxima(omega_w, alpha_w)
    assert np.array([w0r, w0i, wr, wi]) == pytest.approx(
        np.array(refs[0]), abs=0.01)
    assert np.array([I0r, I0i, Ir, Ii]) == pytest.approx(
        np.array(refs[1]), abs=0.05)


def identify_maxima(omega_w, a_w):
    wr, Ir = findpeak(omega_w, a_w.real)
    wi, Ii = findpeak(omega_w, a_w.imag)
    return wr, Ir, wi, Ii


def plot_pol(omega_w, a0_w, a_w):
    import matplotlib.pyplot as plt
    plt.subplot(1, 2, 1)
    plt.plot(omega_w, a0_w.real)
    plt.plot(omega_w, a_w.real)
    plt.subplot(1, 2, 2)
    plt.plot(omega_w, a0_w.imag)
    plt.plot(omega_w, a_w.imag)
    if mpi.world.rank == 0:
        plt.show()



import numpy as np

from ase import Atoms

from gpaw.utilities import h2gpts
from gpaw.grid_descriptor import GridDescriptor


def adjust_cell(atoms: Atoms, border: float,
                h: float = 0.2, idiv: int = 4) -> None:
    """Adjust the cell such that
    1. The vacuum around all atoms is at least border
       in non-periodic directions
    2. The grid spacing chosen by GPAW will be as similar
       as possible to h in all directions
    """
    n_pbc = atoms.pbc.sum()
    if n_pbc == 3:
        return

    pos_ac = atoms.get_positions()
    lowest_c = np.minimum.reduce(pos_ac)
    largest_c = np.maximum.reduce(pos_ac)

    for i, v_c, in enumerate(atoms.cell):
        if (v_c == 0).all():
            assert not atoms.pbc[i]  # pbc with zero cell size make no sense
            atoms.cell[i, i] = 1

    if n_pbc:
        N_c = h2gpts(h, atoms.cell, idiv)
        gd = GridDescriptor(N_c, atoms.cell, atoms.pbc)
        h_c = gd.get_grid_spacings()
        h = 0
        for pbc, h1 in zip(atoms.pbc, h_c):
            if pbc:
                h += h1 / n_pbc

    # the optimal h to be set to non-periodic directions
    h_c = np.array([h, h, h])

    shift_c = np.zeros(3)

    # adjust each cell direction
    for i in range(3):
        if atoms.pbc[i]:
            continue

        # cell direction
        u_c = atoms.cell[i] / np.linalg.norm(atoms.cell[i])

        extension = (largest_c - lowest_c) * u_c
        min_size = extension + 2 * border

        h = h_c[i]
        N = min_size / h
        N = -(N // -idiv) * idiv  # ceil div
        size = N * h

        atoms.cell[i] = size * u_c

        # shift structure to the center
        shift_c += (size - extension) / 2 * u_c
        shift_c -= lowest_c * u_c

    atoms.translate(shift_c)



from __future__ import annotations
from types import SimpleNamespace
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from gpaw.core.atom_arrays import AtomArrays
    from gpaw.core.plane_waves import PWArray
    from gpaw.new.ase_interface import ASECalculator
    from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
    from gpaw.typing import ArrayND


class GSInfo:
    """
    This is the base class for the ground state adapters in the non-linear
    optics module. It is only compatible with GPAW_NEW.

    The class should never be called directly, but should instead be called
    through the CollinearGSInfo or NoncollinearGSInfo classes.

    These subclasses are necessary due to the different ways which the spin
    index is handled in collinear and noncollinear ground state calculations.
    """
    def __init__(self,
                 calc: ASECalculator):
        assert calc.params.mode['name'] == 'pw', \
            'Calculator must be in plane wave mode.'

        dft = calc.dft
        self.nabla_aiiv = [setup.nabla_iiv for setup in dft.setups]

        state = dft.state
        ibzwfs = self.ibzwfs = state.ibzwfs
        if not (ibzwfs.domain_comm.size == 1 and ibzwfs.band_comm.size == 1):
            raise ValueError('Calculator must be initialised with '
                             'only k-point parallelisation.')
        if isinstance(ibzwfs.wfs_qs[0][0].psit_nX, SimpleNamespace):
            raise ValueError('Calculator is missing wfs data. If loading from '
                             'a .gpw file, please recalculate wave functions.')

        density = state.density
        self.collinear = density.collinear
        self.ndensities = density.ndensities

        grid = density.nt_sR.desc
        self.ucvol = np.abs(np.linalg.det(grid.cell))
        self.bzvol = np.abs(np.linalg.det(2 * np.pi * grid.icell))

    def get_plane_wave_coefficients(self,
                                    wfs: PWFDWaveFunctions,
                                    bands: slice,
                                    spin: int) -> tuple[ArrayND, ArrayND]:
        """
        Returns the plane wave coefficients and reciprocal vectors.

        Output is an array with shape (band index, reciprocal vector index)
        """
        psit_nG = wfs.psit_nX[bands]
        G_plus_k_Gv = psit_nG.desc.G_plus_k_Gv
        return G_plus_k_Gv, self._pw_data(psit_nG, spin)

    def get_wave_function_projections(self,
                                      wfs: PWFDWaveFunctions,
                                      bands: slice,
                                      spin: int):
        """
        Returns the projections of the pseudo wfs onto the partial waves.

        Output is a dictionary with atom index keys and array values with
        shape (band index, partial wave index)
        """
        return self._proj_data(wfs.P_ani, bands, spin)

    def get_wfs(self,
                wfs_s: list[PWFDWaveFunctions],
                spin: int) -> PWFDWaveFunctions:
        raise NotImplementedError

    @staticmethod
    def _pw_data(psit: PWArray,
                 spin: int) -> ArrayND:
        raise NotImplementedError

    @staticmethod
    def _proj_data(P: AtomArrays,
                   bands: slice,
                   spin: int) -> dict[int, ArrayND]:
        raise NotImplementedError


class CollinearGSInfo(GSInfo):
    def __init__(self,
                 calc: ASECalculator):
        super().__init__(calc)
        self.ns = self.ndensities

    def get_wfs(self,
                wfs_s: list[PWFDWaveFunctions],
                spin: int) -> PWFDWaveFunctions:
        return wfs_s[spin]

    @staticmethod
    def _pw_data(psit_nG: PWArray,
                 _: int | None = None) -> ArrayND:
        return psit_nG.data

    @staticmethod
    def _proj_data(P_ani: AtomArrays,
                   bands: slice,
                   _: int | None = None) -> dict[int, ArrayND]:
        return {a: P_ni[bands] for a, P_ni in P_ani.items()}


class NoncollinearGSInfo(GSInfo):
    def __init__(self,
                 calc: ASECalculator):
        super().__init__(calc)
        self.ns = 2

    def get_wfs(self,
                wfs_s: list[PWFDWaveFunctions],
                _: int | None = None) -> PWFDWaveFunctions:
        return wfs_s[0]

    @staticmethod
    def _pw_data(psit_nsG: PWArray,
                 spin: int) -> ArrayND:
        return psit_nsG.data[:, spin]

    @staticmethod
    def _proj_data(P_ansi: AtomArrays,
                   bands: slice,
                   spin: int) -> dict[int, ArrayND]:
        return {a: P_nsi[bands, spin] for a, P_nsi in P_ansi.items()}



import numpy as np
from gpaw.response.mpa_interpolation import fit_residue, RESolver
from .mpa_interpolation_scalar import (mpa_R_fit as fit_residue_fortran,
                                       mpa_RE_solver, Xeval)


def test_residues(in_tmp_dir):
    nG = 5
    npols = 10
    Omega_GGp = np.empty((nG, nG, npols), dtype=np.complex128)
    residues_GGp = np.empty((nG, nG, npols), dtype=np.complex128)
    X_GGw = np.empty((nG, nG, 2 * npols), dtype=np.complex128)
    R_fortran_GGp = np.empty((nG, nG, npols), dtype=np.complex128)
    omega_w = np.linspace(0., 5., 2 * npols) + 0.1j

    rng = np.random.default_rng(seed=1)
    for g1 in range(nG):
        for g2 in range(nG):
            Omega_GGp[g1, g2] = rng.random(npols) * 0.05 + 5.5 - 0.01j
            residues_GGp[g1, g2] = rng.random(npols)
            X_GGw[g1, g2] = Xeval(Omega_GGp[g1, g2],
                                  residues_GGp[g1, g2],
                                  omega_w)
            R_fortran_GGp[g1, g2] = fit_residue_fortran(npols, npols, omega_w,
                                                        X_GGw[g1, g2],
                                                        Omega_GGp[g1, g2])

    R_pGG = fit_residue(np.ones((nG, nG)) * npols,
                        omega_w,
                        X_GGw.transpose(2, 0, 1),
                        Omega_GGp.transpose(2, 0, 1))

    R_GGp = R_pGG.transpose(1, 2, 0)

    X_fit_GGw = Xeval(Omega_GGp, R_GGp, omega_w)
    X_fortran_fit_GGw = Xeval(Omega_GGp, R_fortran_GGp, omega_w)
    assert np.allclose(X_fit_GGw, X_fortran_fit_GGw, atol=1e-6)

    if 0:
        from matplotlib import pyplot as plt
        g1, g2 = 0, 0

        plt.plot(omega_w, X_GGw[g1, g2].real, 'k', ls='--')
        plt.plot(omega_w, X_GGw[g1, g2].imag, 'gray', ls='--')

        plt.plot(omega_w, X_fit_GGw[g1, g2].real)
        plt.plot(omega_w, X_fit_GGw[g1, g2].imag)

        plt.plot(omega_w, X_fortran_fit_GGw[g1, g2].real, ls=':')
        plt.plot(omega_w, X_fortran_fit_GGw[g1, g2].imag, ls=':')
        plt.show()


def test_poles(in_tmp_dir):
    nG = 5
    npols = 100
    wmax = 2
    Omega_GGp = np.empty((nG, nG, npols), dtype=np.complex128)
    residues_GGp = np.empty((nG, nG, npols), dtype=np.complex128)

    npols_mpa = 6
    omega_p = np.linspace(0, wmax, npols_mpa)
    omega_w = np.concatenate((omega_p + 0.1j, omega_p + 1.j))

    X_GGw = np.empty((nG, nG, 2 * npols_mpa), dtype=np.complex128)
    E_GGp = np.empty((nG, nG, npols_mpa), dtype=np.complex128)
    R_GGp = np.empty((nG, nG, npols_mpa), dtype=np.complex128)
    E_fortran_GGp = np.empty((nG, nG, npols_mpa), dtype=np.complex128)
    R_fortran_GGp = np.empty((nG, nG, npols_mpa), dtype=np.complex128)

    rng = np.random.default_rng(seed=2)
    for g1 in range(nG):
        for g2 in range(nG):
            Omega_GGp[g1, g2] = rng.normal(1, 0.5, npols) - 0.05j
            residues_GGp[g1, g2] = 0.1 + rng.random(npols)
            X_GGw[g1, g2] = Xeval(Omega_GGp[g1, g2],
                                  residues_GGp[g1, g2],
                                  omega_w)

            R_fortran_GGp[g1, g2], E_fortran_GGp[g1, g2], _, _ = (
                mpa_RE_solver(npols_mpa, omega_w, X_GGw[g1, g2]))
            ind = np.argsort(E_fortran_GGp[g1, g2].real)
            E_fortran_GGp[g1, g2] = E_fortran_GGp[g1, g2, ind]
            R_fortran_GGp[g1, g2] = R_fortran_GGp[g1, g2, ind]

    E_pGG, R_pGG = RESolver(omega_w).solve(X_GGw.transpose(2, 0, 1))

    E_GGp = E_pGG.transpose(1, 2, 0)
    R_GGp = R_pGG.transpose(1, 2, 0)

    assert np.allclose(E_GGp, E_fortran_GGp, rtol=1e-4, atol=1e-6)

    if 0:  # asserting R or X fails due to ill conditioning in np.linalg.solve
        from matplotlib import pyplot as plt

        omega_grid = np.linspace(0., wmax, 100) + 0.01j
        X_fit_GGw = Xeval(E_GGp, R_GGp, omega_grid)
        X_fortran_fit_GGw = Xeval(E_fortran_GGp, R_fortran_GGp, omega_grid)
        a = np.allclose(X_fit_GGw, X_fortran_fit_GGw, rtol=1e-4, atol=1e-6)
        print('assert X', a)

        X_num_GGw = Xeval(Omega_GGp, residues_GGp, omega_grid)
        g1, g2 = 0, 0
        plt.plot(omega_grid.real, X_num_GGw[g1, g2].real, 'k', ls='--')
        plt.plot(omega_grid.real, X_num_GGw[g1, g2].imag, 'gray', ls='--')

        plt.plot(omega_grid.real, X_fit_GGw[g1, g2].real)
        plt.plot(omega_grid.real, X_fit_GGw[g1, g2].imag)
        plt.plot(omega_grid.real, X_fortran_fit_GGw[g1, g2].real, ls=':')
        plt.plot(omega_grid.real, X_fortran_fit_GGw[g1, g2].imag, ls=':')
        plt.show()



from ase import Atoms
from gpaw import GPAW
import pytest


def test_generic_Cl_minus():
    s = Atoms('Cl')
    s.center(vacuum=3)
    c = GPAW(mode='fd',
             xc={'name': 'PBE', 'stencil': 1},
             nbands=-4,
             charge=-1,
             h=0.3)
    s.calc = c

    e = s.get_potential_energy()
    niter = c.get_number_of_iterations()

    print(e, niter)
    energy_tolerance = 0.004
    assert e == pytest.approx(-2.8967, abs=energy_tolerance)



from gpaw import GPAW
from gpaw.eigensolvers.rmmdiis import RMMDIIS
from ase import Atoms
import pytest


def test_eigen_keep_htpsit():
    atoms = Atoms('H')
    atoms.center(3.0)

    convergence = {'eigenstates': 1e-2, 'density': 1e-2}
    base_params = dict(
        mode='fd',
        nbands=2,
        convergence=convergence,
        maxiter=20)
    # Keep htpsit
    calc = GPAW(**base_params, eigensolver=RMMDIIS(keep_htpsit=True))
    atoms.calc = calc
    e0 = atoms.get_potential_energy()

    # Do not keep htpsit
    calc = GPAW(**base_params, eigensolver=RMMDIIS(keep_htpsit=False))
    atoms.calc = calc
    e1 = atoms.get_potential_energy()

    assert e0 == pytest.approx(e1, abs=1e-12)



import pytest

from gpaw import GPAW, FD
from ase import Atoms
import numpy as np
from gpaw.directmin.etdm_fdpw import FDPWETDM
from ase.dft.bandgap import bandgap
from ase.units import Ha


@pytest.mark.later
@pytest.mark.sic
def test_fdsic(in_tmp_dir):
    """
    Test Perdew-Zunger Self-Interaction
    Correction in PW mode using DirectMin
    :param in_tmp_dir:
    :return:
    """

    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * (104.51 + 2.0)
    eps = 0.02
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d + eps, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=4.0)

    calc = GPAW(mode=FD(force_complex_dtype=True),
                h=0.25,
                occupations={'name': 'fixed-uniform'},
                eigensolver=FDPWETDM(
                    functional={'name': 'PZ-SIC',
                                'scaling_factor': (0.5, 0.5)},
                    localizationseed=42,
                    localizationtype='FB_ER',
                    grad_tol_pz_localization=5.0e-3,
                    maxiter_pz_localization=200,
                    converge_unocc=True),
                convergence={'eigenstates': 1e-4},
                mixer={'backend': 'no-mixing'},
                symmetry='off',
                spinpol=True
                )
    H2O.calc = calc
    e = H2O.get_potential_energy()
    f = H2O.get_forces()
    efermi = calc.wfs.fermi_levels[0] * Ha
    gap = bandgap(calc, efermi=efermi)[0]
    e_old = -18.144428

    assert e == pytest.approx(e_old, abs=1e-3)
    # Numeric forces, generated by disabled code below
    f_num = np.array([[2.11270273, 4.89616573e-01, -6.00482442e-04],
                      [-2.15829241, 3.54950512e-01, 1.04418211e-04],
                      [6.67703026e-01, -8.89596180e-01, 8.83126024e-05]])

    assert f == pytest.approx(f_num, abs=5e-2)
    assert gap == pytest.approx(10.215, abs=1e-2)

    numeric = False
    if numeric:
        from ase.calculators.test import numeric_force
        f_num = np.array([[numeric_force(H2O, a, i)
                          for i in range(3)]
                         for a in range(len(H2O))])
        print('Numerical forces')
        print(f_num)
        print(f - f_num, np.abs(f - f_num).max())

    calc.write('h2o.gpw', mode='all')
    from gpaw import restart
    H2O, calc = restart('h2o.gpw', txt='-')
    H2O.positions += 1.0e-6
    e = H2O.get_potential_energy()
    f = H2O.get_forces()
    niter = calc.get_number_of_iterations()
    assert niter == pytest.approx(4, abs=3)
    assert e == pytest.approx(e_old, abs=1e-3)
    assert f == pytest.approx(f_num, abs=5e-2)



from math import pi, sin, cos
import numpy as np

from ase.units import Bohr
import gpaw.cgpaw as cgpaw
from gpaw.poisson import PoissonSolver
from gpaw.pes.state import State
from gpaw.analyse.expandyl import AngularIntegral


class PlaneWave(State):

    """Plane wave state."""

    def get_grid(self, k_c, r0=None):
        if r0 is None:
            r0_c = np.array([0., 0., 0.])
        else:
            r0_c = np.array(r0)
        pw_G = self.gd.zeros(dtype=complex)
        python = True
        python = False
        if python:
            r_c = np.array([0., 0., 0.])
            gd = self.gd
            for i in range(gd.n_c[0]):
                r_c[0] = (gd.beg_c[0] + i) * gd.h_c[0]
                for j in range(gd.n_c[1]):
                    r_c[1] = (gd.beg_c[1] + j) * gd.h_c[1]
                    for k in range(gd.n_c[2]):
                        r_c[2] = (gd.beg_c[2] + k) * gd.h_c[2]
                        phase = np.dot(k_c, r_c - r0_c)
                        pw_G[i, j, k] = cos(phase) + 1j * sin(phase)
        else:
            gd = self.gd
            cgpaw.plane_wave_grid(gd.beg_c, gd.end_c,
                                  gd.h_cv.diagonal().copy(),
                                  k_c, r0_c, pw_G)
        pw_G /= (2 * pi) ** (3. / 2.)
        return pw_G


class ZerothOrder1(State):

    def __init__(self, calculator):
        self.calculator = calculator
        State.__init__(self, calculator.gd)
        self.pw = PlaneWave(self.gd)

        # get effective potential
        hamiltonian = self.calculator.hamiltonian
        if 1:
            self.vt_G = hamiltonian.vt_sG[0]  # XXX treat spin
        else:
            self.vt_G = np.where(hamiltonian.vt_sG[0] > 0,
                                 0.0, hamiltonian.vt_sG[0])
        self.intvt = self.gd.integrate(self.vt_G)
        print('# int(vt_G)=', self.intvt, (self.vt_G > 0).any())

        self.solve()

    def get_grid(self, k_c, r0):
        print('Correction:', self.gd.integrate(self.corrt_G))
        if not hasattr(self, 'written'):
            print('r0', r0, r0 * Bohr)
            dR = 0.3
            AI = AngularIntegral(r0 * Bohr, self.gd, dR=dR)
            r_R = AI.radii()
            psi_R = AI.average(self.corrt_G)
            v_R = AI.average(self.vt_G)
            with open('radial_dR' + str(dR) + '.dat', 'w') as fd:
                print('# R  v(R)    psi(R)', file=fd)
                for r, psi, v in zip(r_R, psi_R, v_R):
                    print(r, psi, v, file=fd)

            self.written = True

        return self.pw.get_grid(k_c, r0) - 1e6 * self.corrt_G

    def solve(self):
        hamiltonian = self.calculator.hamiltonian
        self.poisson = PoissonSolver('fd', nn=hamiltonian.poisson.nn)
        self.poisson.set_grid_descriptor(self.gd)
        self.poisson.initialize()

        corrt_G = self.gd.empty()
        self.poisson.solve(corrt_G, self.vt_G, charge=None)
        corrt_G /= (2 * pi) ** (5. / 2)

        self.corrt_G = corrt_G



import pytest

from gpaw import GPAW, LCAO, restart
from ase import Atoms
import numpy as np
from gpaw.directmin.etdm_lcao import LCAOETDM
from gpaw.directmin.tools import excite
from gpaw.mom import prepare_mom_calculation


@pytest.mark.later
@pytest.mark.sic
def test_mom_lcaosic(in_tmp_dir):
    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * 104.51
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=3.0)

    calc = GPAW(mode=LCAO(force_complex_dtype=True),
                h=0.24,
                basis='sz(dzp)',
                spinpol=True,
                symmetry='off',
                eigensolver='etdm-lcao',
                mixer={'backend': 'no-mixing'},
                occupations={'name': 'fixed-uniform'},
                convergence={'eigenstates': 1e-4}
                )
    H2O.calc = calc
    H2O.get_potential_energy()

    calc.set(eigensolver=LCAOETDM(excited_state=True))
    f_sn = excite(calc, 0, 0, (0, 0))
    prepare_mom_calculation(calc, H2O, f_sn)
    H2O.get_potential_energy()

    test_restart = False
    if test_restart:
        calc.write('h2o.gpw', mode='all')
        H2O, calc = restart('h2o.gpw', txt='-')

    calc.set(eigensolver=LCAOETDM(searchdir_algo={'name': 'l-sr1p'},
                                  linesearch_algo={'name': 'max-step'},
                                  need_init_orbs=False,
                                  localizationtype='PM_PZ',
                                  localizationseed=42,
                                  functional={'name': 'pz-sic',
                                              'scaling_factor': (0.5, 0.5)}),
             convergence={'eigenstates': 1e-2})

    e = H2O.get_potential_energy()
    assert e == pytest.approx(-2.007092, abs=5.0e-3)

    f = H2O.get_forces()

    # Saved analytical forces
    f_old = np.array([[-8.76835930e+00, -1.44190257e+01, 2.53281902e-03],
                      [1.46304433e+01, -1.00089622e+00, -1.03430589e-02],
                      [-4.83182230e+00, 1.51551677e+01, -1.20513818e-02]])

    # Numeric forces, generated by disabled code below
    f_num = np.array([[-8.07002033e+00, -1.51005084e+01, -1.67903017e-03],
                      [1.42573640e+01, -9.80976464e-01, 8.86319871e-05],
                      [-4.93107360e+00, 1.54986283e+01, -3.45104541e-03]])

    numeric = False
    if numeric:
        from ase.calculators.test import numeric_force
        f_num = np.array([[numeric_force(H2O, a, i)
                           for i in range(3)]
                          for a in range(len(H2O))])
        print('Numerical forces')
        print(f_num)
        print(f - f_num, np.abs(f - f_num).max())

    assert f == pytest.approx(f_old, abs=0.5)
    assert f == pytest.approx(f_num, abs=1.0)



from math import log

import numpy as np
import pytest
from ase import Atoms
from ase.db import connect
from ase.io import iread, read, write
from ase.units import Bohr

from gpaw import GPAW, FermiDirac


@pytest.mark.later
def test_generic_hydrogen(in_tmp_dir):
    a = 4.0
    h = 0.2
    hydrogen = Atoms('H',
                     [(a / 2, a / 2, a / 2)],
                     cell=(a, a, a))

    params = dict(mode='fd',
                  h=h,
                  nbands=1,
                  convergence={'energy': 1e-7})
    hydrogen.calc = GPAW(**params,
                         txt='h.txt')
    e1 = hydrogen.get_potential_energy()
    assert e1 == pytest.approx(0.526939, abs=0.001)

    dens = hydrogen.calc.density
    c = dens.gd.find_center(dens.nt_sG[0]) * Bohr
    assert abs(c - a / 2).max() == pytest.approx(0, abs=1e-13)

    kT = 0.001
    hydrogen.calc = GPAW(**params,
                         occupations=FermiDirac(width=kT))
    e2 = hydrogen.get_potential_energy()
    assert e1 == pytest.approx(e2 + log(2) * kT, abs=3.0e-7)

    # Test ase.db a bit:
    for name in ['h.json', 'h.db']:
        con = connect(name, append=False)
        con.write(hydrogen)
        id = con.write(hydrogen, foo='bar', data={'abc': [1, 2, 3]})
        assert id == 2
        assert con.reserve(foo='bar') is None
        row = con.get(foo='bar')
        assert row.energy == e2
        assert sum(row.data.abc) == 6
        del con[1]
        assert con.reserve(x=42) == 3

        write('x' + name, hydrogen)
        write('xx' + name, [hydrogen, hydrogen])

        assert read(name + '@foo=bar')[0].get_potential_energy() == e2
        for n, h in zip([1, 0], iread(name + '@:')):
            assert n == len(h)

    # Test parsing of GPAW's text output:
    h = read('h.txt')
    error = abs(h.calc.get_eigenvalues() -
                hydrogen.calc.get_eigenvalues()).max()
    assert error < 1e-5, error

    # Test get_electrostatic_potential() method
    v = hydrogen.calc.get_electrostatic_potential()
    print(v.shape, np.ptp(v))



from ase import Atoms
from gpaw import GPAW
import pytest


def test_generic_IP_oxygen():
    a = 6.0
    O = Atoms('O',
              [(a / 2, a / 2 + 0.5, a / 2)],
              magmoms=[2],
              pbc=False,
              cell=(a, a + 1, a))
    O.calc = GPAW(mode='fd', gpts=(32, 36, 32), nbands=4)
    e0 = O.get_potential_energy()

    O.calc = GPAW(mode='fd', gpts=(32, 36, 32), nbands=4, charge=1)

    e1 = O.get_potential_energy()

    print(e1 - e0)
    assert abs(e1 - e0 - 13.989) < 0.04

    energy_tolerance = 0.004
    assert e0 == pytest.approx(-1.88477, abs=energy_tolerance)
    assert e1 == pytest.approx(12.11080, abs=energy_tolerance)

    # The first ionization energy for LDA oxygen is from this paper:
    # In-Ho Lee, Richard M. Martin, Phys. Rev. B 56 7197 (1997)



import pytest

from gpaw import GPAW, PW
from ase import Atoms
import numpy as np
from gpaw.directmin.etdm_fdpw import FDPWETDM
from gpaw.directmin.tools import excite
from gpaw.mom import prepare_mom_calculation


@pytest.mark.later
@pytest.mark.sic
def test_mom_pwsic(in_tmp_dir):
    # Water molecule:
    d = 0.9575
    t = np.pi / 180 * 104.51
    H2O = Atoms('OH2',
                positions=[(0, 0, 0),
                           (d, 0, 0),
                           (d * np.cos(t), d * np.sin(t), 0)])
    H2O.center(vacuum=3.0)

    calc = GPAW(mode=PW(300, force_complex_dtype=True),
                spinpol=True,
                symmetry='off',
                eigensolver=FDPWETDM(converge_unocc=True),
                mixer={'backend': 'no-mixing'},
                occupations={'name': 'fixed-uniform'},
                convergence={'eigenstates': 1e-4}
                )
    H2O.calc = calc
    H2O.get_potential_energy()

    calc.set(eigensolver=FDPWETDM(excited_state=True))
    f_sn = excite(calc, 0, 0, (0, 0))
    prepare_mom_calculation(calc, H2O, f_sn)
    H2O.get_potential_energy()

    calc.set(eigensolver=FDPWETDM(
        excited_state=True,
        need_init_orbs=False,
        functional={'name': 'PZ-SIC',
                    'scaling_factor': (0.5, 0.5)},  # SIC/2
        localizationseed=42,
        localizationtype='PM',
        grad_tol_pz_localization=1.0e-2,
        printinnerloop=False,
        grad_tol_inner_loop=1.0e-2),
        convergence={'eigenstates': 1e-3,
                     'density': 1e-3})

    e = H2O.get_potential_energy()
    assert e == pytest.approx(-3.302431, abs=0.2)

    f = H2O.get_forces()

    # Numeric forces, generated by disabled code below
    f_num = np.array([[-2.85022, -3.66863, -0.009059],
                      [3.916824, -0.204146, -0.000065],
                      [-1.204687, 3.822242, 0.000353]])
    assert f == pytest.approx(f_num, abs=0.3)

    numeric = False
    if numeric:
        from ase.calculators.test import numeric_force
        f_num = np.array([[numeric_force(H2O, a, i)
                           for i in range(3)]
                          for a in range(len(H2O))])
        print('Numerical forces')
        print(f_num)
        print(f - f_num, np.abs(f - f_num).max())



from __future__ import annotations

from abc import abstractmethod
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import minimize

from gpaw.response.dyson import HXCScaling, DysonEquation

if TYPE_CHECKING:
    from gpaw.response.chiks import SelfEnhancementCalculator


class GoldstoneScaling(HXCScaling):
    """Scale the Dyson equation to fulfill a Goldstone condition."""

    def _calculate_scaling(self, dyson_equations):
        """Calculate scaling coefficient λ."""
        self.check_descriptors(dyson_equations)

        # Find the frequency to determine the scaling from and identify where
        # the Dyson equation in question is distributed
        wgs = self.find_goldstone_frequency(
            dyson_equations.zd.omega_w)
        wblocks = dyson_equations.zblocks
        rgs, mywgs = wblocks.find_global_index(wgs)

        # Let the rank which holds the Goldstone frequency find and broadcast λ
        lambdbuf = np.empty(1, dtype=float)
        if wblocks.blockcomm.rank == rgs:
            lambdbuf[:] = self.find_goldstone_scaling(dyson_equations[mywgs])
        wblocks.blockcomm.broadcast(lambdbuf, rgs)
        lambd = lambdbuf[0]

        return lambd

    @staticmethod
    def check_descriptors(dyson_equations):
        if not (dyson_equations.qpd.optical_limit and
                dyson_equations.spincomponent in ['+-', '-+']):
            raise ValueError(
                'The Goldstone condition only applies to χ^(+-)(q=0).')

    @abstractmethod
    def find_goldstone_frequency(self, omega_w):
        """Determine frequency index for the Goldstone condition."""

    @abstractmethod
    def find_goldstone_scaling(self, dyson_equation: DysonEquation) -> float:
        """Calculate the Goldstone scaling parameter λ."""


class FMGoldstoneScaling(GoldstoneScaling):
    """Fulfil ferromagnetic Goldstone condition."""

    @staticmethod
    def find_goldstone_frequency(omega_w):
        """Ferromagnetic Goldstone condition is based on χ^(+-)(ω=0)."""
        wgs = np.abs(omega_w).argmin()
        assert abs(omega_w[wgs]) < 1.e-8, \
            "Frequency grid needs to include ω=0."
        return wgs

    @staticmethod
    def find_goldstone_scaling(dyson_equation):
        return find_fm_goldstone_scaling(dyson_equation)


class NewFMGoldstoneScaling(FMGoldstoneScaling):
    """Fulfil Goldstone condition by maximizing a^(+-)(ω=0)."""

    def __init__(self,
                 lambd: float | None = None,
                 m_G: np.ndarray | None = None):
        """Construct the scaling object.

        If the λ-parameter hasn't yet been calculated, the (normalized)
        spin-polarization |m> is needed in order to extract the acoustic
        magnon mode lineshape, a^(+-)(ω=0) = -Im[<m|χ^(+-)(ω=0)|m>]/π.
        """
        super().__init__(lambd=lambd)
        self.m_G = m_G

    @classmethod
    def from_xi_calculator(cls, xi_calc: SelfEnhancementCalculator):
        """Construct scaling object with |m> consistent with a xi_calc."""
        from gpaw.response.localft import (LocalFTCalculator,
                                           add_spin_polarization)
        localft_calc = LocalFTCalculator.from_rshe_parameters(
            xi_calc.gs, xi_calc.context,
            rshelmax=xi_calc.rshelmax, rshewmin=xi_calc.rshewmin)
        qpd = xi_calc.get_pw_descriptor(q_c=[0., 0., 0.])
        nz_G = localft_calc(qpd, add_spin_polarization)
        return cls(m_G=nz_G / np.linalg.norm(nz_G))

    def find_goldstone_scaling(self, dyson_equation):
        assert self.m_G is not None, \
            'Please supply spin-polarization to calculate λ'

        def acoustic_antispectrum(lambd):
            """Calculate -a^(+-)(ω=0)."""
            return - calculate_acoustic_spectrum(
                lambd, dyson_equation, self.m_G)

        # Maximize a^(+-)(ω=0)
        res = minimize(acoustic_antispectrum, x0=[1.], bounds=[(0.1, 10.)])
        assert res.success
        return res.x[0]


class AFMGoldstoneScaling(GoldstoneScaling):
    """Fulfil antiferromagnetic Goldstone condition."""

    @staticmethod
    def find_goldstone_frequency(omega_w):
        """Antiferromagnetic Goldstone condition is based on ω->0^+."""
        # Set ω<=0. to np.inf
        omega1_w = np.where(omega_w < 1.e-8, np.inf, omega_w)
        # Sort for the two smallest positive frequencies
        omega2_w = np.partition(omega1_w, 1)
        # Find original index of second smallest positive frequency
        wgs = np.abs(omega_w - omega2_w[1]).argmin()
        return wgs

    @staticmethod
    def find_goldstone_scaling(dyson_equation):
        return find_afm_goldstone_scaling(dyson_equation)


def find_fm_goldstone_scaling(dyson_equation):
    """Find goldstone scaling of the kernel by ensuring that the
    macroscopic inverse enhancement function has a root in (q=0, omega=0)."""
    fnct = partial(calculate_macroscopic_kappa,
                   dyson_equation=dyson_equation)

    def is_converged(kappaM):
        return abs(kappaM) < 1e-7

    return find_root(fnct, is_converged)


def find_afm_goldstone_scaling(dyson_equation):
    """Find goldstone scaling of the kernel by ensuring that the
    macroscopic magnon spectrum vanishes at q=0. for finite frequencies."""
    fnct = partial(calculate_macroscopic_spectrum,
                   dyson_equation=dyson_equation)

    def is_converged(SM):
        # We want the macroscopic spectrum to be strictly positive
        return 0. < SM < 1.e-7

    return find_root(fnct, is_converged)


def find_root(fnct, is_converged):
    """Find the root f(λ)=0, where the scaling parameter λ~1.

    |f(λ)| is minimized iteratively, assuming that f(λ) is continuous and
    monotonically decreasing with λ for λ∊]0.1, 10[.
    """
    lambd = 1.  # initial guess for the scaling parameter
    value = fnct(lambd)
    lambd_incr = 0.1 * np.sign(value)  # Increase λ to decrease f(λ)
    while not is_converged(value) or abs(lambd_incr) > 1.e-7:
        # Update λ
        lambd += lambd_incr
        if lambd <= 0.1 or lambd >= 10.:
            raise Exception(f'Found an invalid λ-value of {lambd:.4f}')
        # Update value and refine increment, if we have passed f(λ)=0
        value = fnct(lambd)
        if np.sign(value) != np.sign(lambd_incr):
            lambd_incr *= -0.2
    return lambd


def calculate_macroscopic_kappa(lambd, dyson_equation):
    """Invert dyson equation and calculate the inverse enhancement function."""
    chi_GG = dyson_equation.invert(lambd=lambd)
    return (dyson_equation.chiks_GG[0, 0] / chi_GG[0, 0]).real


def calculate_macroscopic_spectrum(lambd, dyson_equation):
    """Invert dyson equation and extract the macroscopic spectrum."""
    chi_GG = dyson_equation.invert(lambd=lambd)
    return - chi_GG[0, 0].imag / np.pi


def calculate_acoustic_spectrum(lambd, dyson_equation, m_G):
    """Invert the dyson equation and extract the acoustic spectrum."""
    chi_GG = dyson_equation.invert(lambd=lambd)
    chi_projection = np.conj(m_G) @ chi_GG @ m_G
    return - chi_projection.imag / np.pi



from gpaw.directmin.fdpw.pz_localization import PZLocalization
from gpaw.directmin.fdpw.etdm_inner_loop import ETDMInnerLoop
import numpy as np


class FDPWETDMLocalize:

    def __init__(self, obj_f, wfs, maxiter=300, g_tol=1.0e-3, randval=0.01):

        self.obj_f = obj_f
        self.randval = randval

        if obj_f.name == 'Zero':
            self.iloop = ETDMInnerLoop(
                obj_f, wfs, 'all', 5.0e-4, maxiter,
                g_tol=g_tol, useprec=True)
        else:
            self.iloop = PZLocalization(
                obj_f, wfs, maxiter=maxiter,
                g_tol=g_tol)

    def run(self, wfs, dens, ham=None, rewritepsi=True, log=None, seed=None):

        wfs.timer.start('Inner loop')

        if self.obj_f.name == 'Zero':
            etotal, counter = self.iloop.run(wfs, dens, log, 0, ham=ham)
        else:
            counter = self.iloop.run(0.0, wfs, dens, log, 0,
                                     randvalue=self.randval, seed=seed)

        if rewritepsi:
            for kpt in wfs.kpt_u:
                k = self.iloop.n_kps * kpt.s + kpt.q
                dim1 = self.iloop.U_k[k].shape[0]
                kpt.psit_nG[:dim1] = np.tensordot(
                    self.iloop.U_k[k].T, kpt.psit_nG[:dim1], axes=1)
                wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)

        wfs.timer.stop('Inner loop')

        return counter



import pytest
from contextlib import contextmanager

from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2


@contextmanager
def no_error():
    yield


pytestmark = pytest.mark.usefixtures('module_tmp_path')


@pytest.mark.lrtddft
@pytest.mark.parametrize('max_energy_diff, expectation', [
    (15, pytest.raises(RuntimeError, match=r'.* HOMO \(n=3\) .*')),
    (10, no_error())])
def test_max_energy_diff(gpw_files, max_energy_diff, expectation,
                         in_tmp_dir):
    calc = GPAW(gpw_files['h20_lr2_nbands6'])
    with expectation:
        LrTDDFT2('lr2', calc, fxc='LDA', max_energy_diff=max_energy_diff)


@pytest.mark.lrtddft
@pytest.mark.parametrize('min_occ', [None, 0])
@pytest.mark.parametrize('min_unocc', [None, 0])
@pytest.mark.parametrize('max_occ', [None, 5])
@pytest.mark.parametrize('max_unocc', [None, 5])
def test_indices_with_max_energy_diff(gpw_files, min_occ, min_unocc,
                                      max_occ, max_unocc, in_tmp_dir):
    calc = GPAW(gpw_files['h20_lr2_nbands6'])

    if (min_occ is None or min_unocc is None
        or max_occ is None or max_unocc is None):
        expectation = pytest.raises(RuntimeError)
    else:
        expectation = no_error()

    with expectation:
        LrTDDFT2('lr2', calc, fxc='LDA', max_energy_diff=15,
                 min_occ=min_occ, min_unocc=min_unocc,
                 max_occ=max_occ, max_unocc=max_unocc)


@pytest.mark.lrtddft
@pytest.mark.parametrize('min_occ', [None, 0])
@pytest.mark.parametrize('min_unocc', [None, 0])
@pytest.mark.parametrize('max_occ', [None, 5])
@pytest.mark.parametrize('max_unocc', [None, 5])
def test_indices(gpw_files, min_occ, min_unocc,
                 max_occ, max_unocc, in_tmp_dir):
    calc = GPAW(gpw_files['h20_lr2_nbands6'])
    LrTDDFT2('lr2', calc, fxc='LDA',
             min_occ=min_occ, min_unocc=min_unocc,
             max_occ=max_occ, max_unocc=max_unocc)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass

import numpy as np

from gpaw.response import timer
from gpaw.response.pair_functions import Chi


class HXCScaling(ABC):
    """Helper for scaling the hxc contribution to Dyson equations."""

    def __init__(self, lambd=None):
        self._lambd = lambd

    @property
    def lambd(self):
        return self._lambd

    def calculate_scaling(self, dyson_equations):
        self._lambd = self._calculate_scaling(dyson_equations)

    @abstractmethod
    def _calculate_scaling(self, dyson_equations: DysonEquations) -> float:
        """Calculate hxc scaling coefficient."""


class PWKernel(ABC):
    """Hartree, exchange and/correleation kernels in a plane-wave basis."""

    @property
    def nG(self):
        return self.get_number_of_plane_waves()

    def add_to(self, x_GG):
        assert x_GG.shape == (self.nG, self.nG)
        self._add_to(x_GG)

    @abstractmethod
    def get_number_of_plane_waves(self):
        """Return the size of the plane-wave basis."""

    @abstractmethod
    def _add_to(self, x_GG):
        """Add plane-wave kernel to an input array."""


class NoKernel(PWKernel):
    """A plane-wave kernel equal to zero."""

    def __init__(self, *, nG):
        self._nG = nG

    @classmethod
    def from_qpd(cls, qpd):
        qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
        return cls(nG=qG_Gv.shape[0])

    def get_number_of_plane_waves(self):
        return self._nG

    def _add_to(self, x_GG):
        pass


@dataclass
class HXCKernel:
    """Hartree-exchange-correlation kernel in a plane-wave basis."""
    hartree_kernel: PWKernel
    xc_kernel: PWKernel

    def __post_init__(self):
        assert self.hartree_kernel.nG == self.xc_kernel.nG
        self.nG = self.hartree_kernel.nG

    def get_Khxc_GG(self):
        """Hartree-exchange-correlation kernel."""
        # Allocate array
        Khxc_GG = np.zeros((self.nG, self.nG), dtype=complex)
        self.hartree_kernel.add_to(Khxc_GG)
        self.xc_kernel.add_to(Khxc_GG)
        return Khxc_GG


class DysonSolver:
    """Class for inversion of Dyson-like equations."""

    def __init__(self, context):
        self.context = context

    @timer('Invert Dyson-like equations')
    def __call__(self, chiks: Chi, self_interaction: HXCKernel | Chi,
                 hxc_scaling: HXCScaling | None = None) -> Chi:
        """Solve the dyson equation and return the many-body susceptibility."""
        dyson_equations = self.get_dyson_equations(chiks, self_interaction)
        if hxc_scaling:
            if not hxc_scaling.lambd:  # calculate, if it hasn't been already
                hxc_scaling.calculate_scaling(dyson_equations)
            lambd = hxc_scaling.lambd
            self.context.print(r'Rescaling the self-enhancement function by a '
                               f'factor of λ={lambd}')
        self.context.print('Inverting Dyson-like equations')
        return dyson_equations.invert(hxc_scaling=hxc_scaling)

    @staticmethod
    def get_dyson_equations(chiks, self_interaction):
        if isinstance(self_interaction, HXCKernel):
            return DysonEquationsWithKernel(chiks, hxc_kernel=self_interaction)
        elif isinstance(self_interaction, Chi):
            return DysonEquationsWithXi(chiks, xi=self_interaction)
        else:
            raise ValueError(
                f'Invalid encoding of the self-interaction {self_interaction}')


class DysonEquations(Sequence):
    """Sequence of Dyson-like equations at different complex frequencies z."""

    def __init__(self, chiks: Chi):
        assert chiks.distribution == 'zGG' and\
            chiks.blockdist.fully_block_distributed, \
            "chiks' frequencies need to be fully distributed over world"
        self.chiks = chiks
        # Inherit basic descriptors from chiks
        self.qpd = chiks.qpd
        self.zd = chiks.zd
        self.zblocks = chiks.blocks1d
        self.spincomponent = chiks.spincomponent

    def __len__(self):
        return self.zblocks.nlocal

    def invert(self, hxc_scaling: HXCScaling | None = None) -> Chi:
        """Invert Dyson equations to obtain χ(z)."""
        # Scaling coefficient of the self-enhancement function
        lambd = hxc_scaling.lambd if hxc_scaling else None
        chi = self.chiks.new()
        for z, dyson_equation in enumerate(self):
            chi.array[z] = dyson_equation.invert(lambd=lambd)
        return chi


class DysonEquationsWithKernel(DysonEquations):
    def __init__(self, chiks: Chi, *, hxc_kernel: HXCKernel):
        # Check compatibility
        nG = hxc_kernel.nG
        assert chiks.array.shape[1:] == (nG, nG)
        # Initialize
        super().__init__(chiks)
        self.Khxc_GG = hxc_kernel.get_Khxc_GG()

    def __getitem__(self, z):
        chiks_GG = self.chiks.array[z]
        xi_GG = chiks_GG @ self.Khxc_GG
        return DysonEquation(chiks_GG, xi_GG)


class DysonEquationsWithXi(DysonEquations):
    def __init__(self, chiks: Chi, *, xi: Chi):
        # Check compatibility
        assert xi.distribution == 'zGG' and \
            xi.blockdist.fully_block_distributed
        assert chiks.spincomponent == xi.spincomponent
        assert np.allclose(chiks.zd.hz_z, xi.zd.hz_z)
        assert np.allclose(chiks.qpd.q_c, xi.qpd.q_c)
        # Initialize
        super().__init__(chiks)
        self.xi = xi

    def __getitem__(self, z):
        return DysonEquation(self.chiks.array[z], self.xi.array[z])


class DysonEquation:
    """Dyson equation at wave vector q and frequency z.

    The Dyson equation is given in plane-wave components as

    χ(q,z) = χ_KS(q,z) + Ξ(q,z) χ(q,z),

    where the self-enhancement function encodes the electron correlations
    induced by by the effective (Hartree-exchange-correlation) interaction:

    Ξ(q,z) = χ_KS(q,z) K_hxc(q,z)

    See [to be published] for more information.
    """

    def __init__(self, chiks_GG, xi_GG):
        self.nG = chiks_GG.shape[0]
        self.chiks_GG = chiks_GG
        self.xi_GG = xi_GG

    def invert(self, lambd: float | None = None):
        """Invert the Dyson equation (with or without a rescaling of Ξ).

        χ(q,z) = [1 - λ Ξ(q,z)]^(-1) χ_KS(q,z)
        """
        if lambd is None:
            lambd = 1.  # no rescaling
        enhancement_GG = np.linalg.inv(np.eye(self.nG) - lambd * self.xi_GG)
        return enhancement_GG @ self.chiks_GG



import pytest
from ase.atoms import Atoms

from gpaw import GPAW, PW


@pytest.mark.later
def test_xc_qna_spinpol(in_tmp_dir):
    QNA = {'alpha': 2.0,
           'name': 'QNA',
           'orbital_dependent': False,
           'parameters': {'H': (0.1485, 0.005)},
           'setup_name': 'PBE',
           'type': 'qna-gga'}

    atoms = Atoms('H', pbc=True, cell=[5, 5, 5])
    atoms.set_initial_magnetic_moments([1])

    calc = GPAW(mode=PW(400),
                kpts=(1, 1, 1),
                xc=QNA,
                parallel={'domain': 1},
                txt='qna_spinpol.txt')

    atoms.calc = calc
    atoms.get_potential_energy()
    magmoms = atoms.get_magnetic_moments()

    tol = 0.003
    assert 0.25374 == pytest.approx(magmoms[0], abs=tol)



import pytest

from ase import Atoms

from gpaw import GPAW
from gpaw.mpi import size


def test_noncollinear_o2(in_tmp_dir, gpaw_new):
    if size > 2:
        pytest.skip('mpi world size >2')
    if not gpaw_new:
        pytest.skip('fatal crash with old code')

    a = Atoms('OO', [[0, 0, 0], [0, 0, 1.1]], magmoms=[1, 1], pbc=(1, 0, 0))
    a.center(vacuum=2.5)
    a.calc = GPAW(mode='pw',
                  kpts=(2, 1, 1))
    f0 = a.get_forces()
    e0 = a.calc.get_eigenvalues(0, 0)[5]
    p0 = a.calc.get_pseudo_wave_function(5, periodic=True)

    a.calc = GPAW(mode='pw',
                  kpts=(2, 1, 1),
                  symmetry='off',
                  experimental={'magmoms': [[0, 0.5, 0.5], [0, 0, 1]]})
    f = a.get_forces()
    e = a.calc.get_eigenvalues(0, 0)[10]
    p = a.calc.get_pseudo_wave_function(10, periodic=True)

    assert abs(f - f0).max() < 0.01
    assert e == pytest.approx(e0, abs=0.002)
    assert abs(p0)**2 == pytest.approx((abs(p)**2).sum(axis=0), abs=1e-4)

    m1_v = a.calc.get_non_collinear_magnetic_moment()
    m1_av = a.calc.get_non_collinear_magnetic_moments()
    a.calc.write('o2.gpw')
    a.calc.write('o2w.gpw', 'all')
    calc = GPAW('o2w.gpw')
    m2_v = calc.get_non_collinear_magnetic_moment()
    m2_av = calc.get_non_collinear_magnetic_moments()
    m3_v, m3_av = calc.dft.magmoms()  # recompute
    assert m1_v == pytest.approx(m2_v)
    assert m1_av == pytest.approx(m2_av)
    assert m1_v == pytest.approx(m3_v)
    assert m1_av == pytest.approx(m3_av)

    p = calc.get_pseudo_wave_function(10, periodic=True)
    assert abs(p0)**2 == pytest.approx((abs(p)**2).sum(axis=0), abs=1e-4)

    if gpaw_new:
        n_sR = calc.dft.densities().all_electron_densities()
        print(n_sR)



from __future__ import annotations

from functools import cached_property
from typing import Any, Union

import numpy as np
from ase import Atoms
from ase.units import Bohr, Ha

from gpaw.core import UGDesc
from gpaw.core.atom_arrays import AtomDistribution
from gpaw.densities import Densities
from gpaw.electrostatic_potential import ElectrostaticPotential
from gpaw.gpu import as_np
from gpaw.mpi import broadcast_float, world
from gpaw.new import trace, zips
from gpaw.new.density import Density
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.input_parameters import InputParameters
from gpaw.new.logger import Logger
from gpaw.new.potential import Potential
from gpaw.new.scf import SCFLoop
from gpaw.setup import Setups
from gpaw.typing import Array1D, Array2D
from gpaw.utilities import (check_atoms_too_close,
                            check_atoms_too_close_to_boundary)
from gpaw.utilities.partition import AtomPartition


class ReuseWaveFunctionsError(Exception):
    """Reusing the old wave functions after cell-change failed.

    Most likekly, the number of k-points changed.
    """


class NonsenseError(Exception):
    """Operation doesn't make sense."""


class CalculationModeError(Exception):
    """Calculation mode does not match what is expected from a given method.

    For example, if a method only works in collinear mode and receives a
    calculator in non-collinear mode, this exception should be raised.
    """


units = {'energy': Ha,
         'free_energy': Ha,
         'forces': Ha / Bohr,
         'stress': Ha / Bohr**3,
         'dipole': Bohr,
         'magmom': 1.0,
         'magmoms': 1.0,
         'non_collinear_magmom': 1.0,
         'non_collinear_magmoms': 1.0}


class DFTState:
    def __init__(self,
                 ibzwfs: IBZWaveFunctions,
                 density: Density,
                 potential: Potential):
        """State of a Kohn-Sham calculation."""
        self.ibzwfs = ibzwfs
        self.density = density
        self.potential = potential

    def __repr__(self):
        return (f'DFTState({self.ibzwfs!r}, '
                f'{self.density!r}, {self.potential!r})')

    def __str__(self):
        return f'{self.ibzwfs}\n{self.density}\n{self.potential}'

    def move(self, fracpos_ac, atomdist):
        self.ibzwfs.move(fracpos_ac, atomdist)
        self.density.move(fracpos_ac, atomdist)
        self.potential.move(atomdist)


class DFTCalculation:
    def __init__(self,
                 state: DFTState,
                 setups: Setups,
                 scf_loop: SCFLoop,
                 pot_calc,
                 log: Logger):
        self.state = state
        self.setups = setups
        self.scf_loop = scf_loop
        self.pot_calc = pot_calc
        self.log = log
        self.comm = log.comm

        self.results: dict[str, Any] = {}
        self.fracpos_ac = self.pot_calc.fracpos_ac

    @classmethod
    def from_parameters(cls,
                        atoms: Atoms,
                        params: Union[dict, InputParameters],
                        comm=None,
                        log=None,
                        builder=None) -> DFTCalculation:
        """Create DFTCalculation object from parameters and atoms."""
        from gpaw.new.builder import builder as create_builder

        check_atoms_too_close(atoms)
        check_atoms_too_close_to_boundary(atoms)

        if params is None:
            params = {}
        if isinstance(params, dict):
            params = InputParameters(params)

        if not isinstance(log, Logger):
            log = Logger(log, comm or world)

        builder = builder or create_builder(atoms, params, log.comm)

        basis_set = builder.create_basis_set()

        density = builder.density_from_superposition(basis_set)
        density.normalize()

        # The SCF-loop has a Hamiltonian that has an fft-plan that is
        # cached for later use, so best to create the SCF-loop first
        # FIX this!
        scf_loop = builder.create_scf_loop()

        pot_calc = builder.create_potential_calculator()
        potential, _ = pot_calc.calculate_without_orbitals(
            density, kpt_band_comm=builder.communicators['D'])
        ibzwfs = builder.create_ibz_wave_functions(
            basis_set, potential, log=log)

        if ibzwfs.wfs_qs[0][0]._eig_n is not None:
            ibzwfs.calculate_occs(scf_loop.occ_calc)

        state = DFTState(ibzwfs, density, potential)

        write_atoms(atoms, builder.initial_magmom_av, builder.grid, log)
        log(state)
        log(builder.setups)
        log(scf_loop)
        log(pot_calc)

        return cls(state, builder.setups, scf_loop, pot_calc, log)

    def move_atoms(self, atoms) -> DFTCalculation:
        check_atoms_too_close(atoms)

        self.fracpos_ac = np.ascontiguousarray(atoms.get_scaled_positions())
        self.comm.broadcast(self.fracpos_ac, 0)

        atomdist = self.state.density.D_asii.layout.atomdist
        grid = self.state.density.nt_sR.desc
        rank_a = grid.ranks_from_fractional_positions(self.fracpos_ac)
        atomdist = AtomDistribution(rank_a, atomdist.comm)

        self.pot_calc.move(self.fracpos_ac, atomdist)
        self.state.move(self.fracpos_ac, atomdist)

        mm_av = self.results['non_collinear_magmoms']
        write_atoms(atoms, mm_av, self.state.density.nt_sR.desc, self.log)

        self.results = {}

        return self

    def iconverge(self, convergence=None, maxiter=None, calculate_forces=None):
        self.state.ibzwfs.make_sure_wfs_are_read_from_gpw_file()
        yield from self.scf_loop.iterate(self.state,
                                         self.pot_calc,
                                         convergence,
                                         maxiter,
                                         calculate_forces,
                                         log=self.log)

    @trace
    def converge(self,
                 convergence=None,
                 maxiter=None,
                 steps=99999999999999999,
                 calculate_forces=None):
        """Converge to self-consistent solution of Kohn-Sham equation."""
        for step, _ in enumerate(self.iconverge(convergence,
                                                maxiter,
                                                calculate_forces),
                                 start=1):
            if step == steps:
                break
        else:  # no break
            self.log(scf_steps=step)

    def energies(self):
        energies = combine_energies(self.state.potential, self.state.ibzwfs)

        self.log('Energy contributions relative to reference atoms:',
                 f'(reference = {self.setups.Eref * Ha:.6f})\n')

        for name, e in energies.items():
            if not name.startswith('total') and name != 'stress':
                self.log(f'{name + ":":10}   {e * Ha:14.6f}')
        total_free = energies['total_free']
        total_extrapolated = energies['total_extrapolated']
        self.log('----------------------------')
        self.log(f'Free energy: {total_free * Ha:14.6f}')
        self.log(f'Extrapolated:{total_extrapolated * Ha:14.6f}\n')

        total_free = broadcast_float(total_free, self.comm)
        total_extrapolated = broadcast_float(total_extrapolated, self.comm)

        self.results['free_energy'] = total_free
        self.results['energy'] = total_extrapolated

    def dipole(self):
        if 'dipole' in self.results:
            return
        dipole_v = self.state.density.calculate_dipole_moment(self.fracpos_ac)
        x, y, z = dipole_v * Bohr
        self.log(f'dipole moment: [{x:.6f}, {y:.6f}, {z:.6f}]  # |e|*Ang\n')
        self.results['dipole'] = dipole_v

    def magmoms(self) -> tuple[Array1D, Array2D]:
        mm_v, mm_av = self.state.density.calculate_magnetic_moments()
        self.results['magmom'] = mm_v[2]
        self.results['magmoms'] = mm_av[:, 2].copy()
        self.results['non_collinear_magmoms'] = mm_av
        self.results['non_collinear_magmom'] = mm_v

        if self.state.density.ncomponents > 1:
            x, y, z = mm_v
            self.log(f'total magnetic moment: [{x:.6f}, {y:.6f}, {z:.6f}]\n')
            self.log('local magnetic moments: [')
            for a, (setup, m_v) in enumerate(zips(self.setups, mm_av)):
                x, y, z = m_v
                c = ',' if a < len(mm_av) - 1 else ']'
                self.log(f'  [{x:9.6f}, {y:9.6f}, {z:9.6f}]{c}'
                         f'  # {setup.symbol:2} {a}')
            self.log()
        return mm_v, mm_av

    def forces(self, silent=False):
        """Calculate atomic forces."""
        if 'forces' not in self.results or silent:
            self._calculate_forces()

            if silent:
                return
            self.log('\nforces: [  # eV/Ang')
            F_av = self.results['forces'] * (Ha / Bohr)
            for a, setup in enumerate(self.setups):
                x, y, z = F_av[a]
                c = ',' if a < len(F_av) - 1 else ']'
                self.log(f'  [{x:9.3f}, {y:9.3f}, {z:9.3f}]{c}'
                         f'  # {setup.symbol:2} {a}')

    def _calculate_forces(self):
        xc = self.pot_calc.xc
        assert not hasattr(xc.xc, 'setup_force_corrections')

        # Force from projector functions (and basis set):
        F_av = self.state.ibzwfs.forces(self.state.potential)

        pot_calc = self.pot_calc
        Fcc_avL, Fnct_av, Ftauct_av, Fvbar_av = pot_calc.force_contributions(
            self.state)

        # Force from compensation charges:
        ccc_aL = \
            self.state.density.calculate_compensation_charge_coefficients()
        for a, dF_vL in Fcc_avL.items():
            F_av[a] += dF_vL @ ccc_aL[a]

        # Force from smooth core charge:
        for a, dF_v in Fnct_av.items():
            F_av[a] += dF_v[:, 0]

        if Ftauct_av is not None:
            # Force from smooth core ked:
            for a, dF_v in Ftauct_av.items():
                F_av[a] += dF_v[:, 0]

        # Force from zero potential:
        for a, dF_v in Fvbar_av.items():
            F_av[a] += dF_v[:, 0]

        F_av = as_np(F_av)

        domain_comm = ccc_aL.layout.atomdist.comm
        domain_comm.sum(F_av)

        F_av = self.state.ibzwfs.ibz.symmetries.symmetrize_forces(F_av)
        self.comm.broadcast(F_av, 0)
        self.results['forces'] = F_av

    def stress(self) -> None:
        if 'stress' in self.results:
            return
        stress_vv = self.pot_calc.stress(self.state)
        self.log('\nstress tensor: [  # eV/Ang^3')
        for (x, y, z), c in zips(stress_vv * (Ha / Bohr**3), ',,]'):
            self.log(f'  [{x:13.6f}, {y:13.6f}, {z:13.6f}]{c}')
        self.results['stress'] = stress_vv.flat[[0, 4, 8, 5, 2, 1]]

    def write_converged(self) -> None:
        self.state.ibzwfs.write_summary(self.log)
        vacuum_level = self.state.potential.get_vacuum_level()
        if not np.isnan(vacuum_level):
            self.log(f'vacuum-level: {vacuum_level:.3f}  # V')
            try:
                wf1, wf2 = self.workfunctions(vacuum_level=vacuum_level)
            except NonsenseError:
                pass
            else:
                self.log(f'Workfunctions: {wf1:.3f}, {wf2:.3f}  # eV')
        self.log.fd.flush()

    def workfunctions(self,
                      *,
                      vacuum_level: float | None = None
                      ) -> tuple[float, float]:
        if vacuum_level is None:
            vacuum_level = self.state.potential.get_vacuum_level()
        if np.isnan(vacuum_level):
            raise NonsenseError('No vacuum')
        try:
            correction = self.pot_calc.poisson_solver.dipole_layer_correction()
        except NotImplementedError:
            raise NonsenseError('No dipole layer')
        correction *= Ha
        fermi_level = self.state.ibzwfs.fermi_level * Ha
        wf = vacuum_level - fermi_level
        return wf - correction, wf + correction

    def electrostatic_potential(self) -> ElectrostaticPotential:
        return ElectrostaticPotential.from_calculation(self)

    def densities(self) -> Densities:
        return Densities.from_calculation(self)

    @cached_property
    def _atom_partition(self):
        # Backwards compatibility helper
        atomdist = self.state.density.D_asii.layout.atomdist
        return AtomPartition(atomdist.comm, atomdist.rank_a)

    def new(self,
            atoms: Atoms,
            params: InputParameters,
            log=None) -> DFTCalculation:
        """Create new DFTCalculation object."""
        from gpaw.new.builder import builder as create_builder

        if params.mode['name'] != 'pw':
            raise ReuseWaveFunctionsError

        ibzwfs = self.state.ibzwfs
        if ibzwfs.domain_comm.size != 1:
            raise ReuseWaveFunctionsError

        if not self.state.density.nt_sR.desc.pbc_c.all():
            raise ReuseWaveFunctionsError

        check_atoms_too_close(atoms)
        check_atoms_too_close_to_boundary(atoms)

        builder = create_builder(atoms, params, self.comm)

        kpt_kc = builder.ibz.kpt_kc
        old_kpt_kc = ibzwfs.ibz.kpt_kc
        if len(kpt_kc) != len(old_kpt_kc):
            raise ReuseWaveFunctionsError
        if abs(kpt_kc - old_kpt_kc).max() > 1e-9:
            raise ReuseWaveFunctionsError

        log('# Interpolating wave functions to new cell')

        density = self.state.density.new(builder.grid,
                                         builder.interpolation_desc,
                                         builder.fracpos_ac,
                                         builder.atomdist)
        density.normalize()

        # Make sure all have exactly the same density.
        # Not quite sure it is needed???
        # At the moment we skip it on GPU's because it doesn't
        # work!
        if density.nt_sR.xp is np:
            self.comm.broadcast(density.nt_sR.data, 0)

        scf_loop = builder.create_scf_loop()
        pot_calc = builder.create_potential_calculator()
        potential, _ = pot_calc.calculate(density)

        old_ibzwfs = ibzwfs

        def create_wfs(spin, q, k, kpt_c, weight):
            wfs = old_ibzwfs.wfs_qs[q][spin]
            return wfs.morph(
                builder.wf_desc,
                builder.fracpos_ac,
                builder.atomdist)

        ibzwfs = ibzwfs.create(
            ibz=builder.ibz,
            nelectrons=old_ibzwfs.nelectrons,
            ncomponents=old_ibzwfs.ncomponents,
            create_wfs_func=create_wfs,
            kpt_comm=old_ibzwfs.kpt_comm,
            kpt_band_comm=old_ibzwfs.kpt_band_comm,
            comm=self.comm)

        state = DFTState(ibzwfs, density, potential)

        write_atoms(atoms, builder.initial_magmom_av, builder.grid, log)
        log(state)
        log(builder.setups)
        log(scf_loop)
        log(pot_calc)

        return DFTCalculation(
            state, builder.setups, scf_loop, pot_calc, log)


def combine_energies(potential: Potential,
                     ibzwfs: IBZWaveFunctions) -> dict[str, float]:
    """Add up energy contributions."""
    energies = potential.energies.copy()
    energies.pop('stress', 0.0)
    energies['kinetic'] += ibzwfs.energies['band']
    energies['kinetic'] += ibzwfs.energies.get('exx_kinetic', 0.0)
    energies['xc'] += (ibzwfs.energies.get('exx_vv', 0.0) +
                       ibzwfs.energies.get('exx_vc', 0.0) +
                       ibzwfs.energies.get('exx_cc', 0.0))
    energies['entropy'] = ibzwfs.energies['entropy']
    energies['total_free'] = sum(energies.values())
    energies['total_extrapolated'] = (energies['total_free'] +
                                      ibzwfs.energies['extrapolation'])
    return energies


def write_atoms(atoms: Atoms,
                magmom_av: Array2D,
                grid: UGDesc,
                log) -> None:
    from gpaw.output import print_cell, print_positions
    print_positions(atoms, log, magmom_av)
    print_cell(grid._gd, grid.pbc, log)



import pytest
from gpaw.xc import XC


@pytest.mark.gpu
@pytest.mark.parametrize('nspins', [1, 2])
def test_gpu_pbe(nspins, gpu):
    import cupy as cp
    from gpaw.cgpaw import evaluate_pbe_gpu
    ng = 10000
    n_sg = cp.exp(cp.log(10) * 5 * (cp.random.rand(nspins, ng) - 0.5))
    sigma_xg = cp.exp(cp.log(10) * 5 * (
                      cp.random.rand(2 * nspins - 1, ng) - 0.5))
    dedsigma_xg = cp.zeros_like(sigma_xg)
    cp.cuda.runtime.deviceSynchronize()
    v_sg = cp.zeros_like(n_sg)
    e_g = cp.zeros((ng,))

    cpun_sg = cp.asnumpy(n_sg)
    cpuv_sg = cp.asnumpy(v_sg)
    cpue_g = cp.asnumpy(e_g)
    cpusigma_xg = cp.asnumpy(sigma_xg)
    cpudedsigma_xg = cp.asnumpy(dedsigma_xg)
    xc = XC('PBE')
    import time
    start = time.time()
    xc.kernel.calculate(cpue_g, cpun_sg, cpuv_sg, cpusigma_xg, cpudedsigma_xg)
    stop = time.time()
    cpu = stop - start
    print('CPU took', stop - start)

    cp.cuda.runtime.deviceSynchronize()
    start = time.time()
    evaluate_pbe_gpu(n_sg, v_sg, e_g, sigma_xg, dedsigma_xg)
    cp.cuda.runtime.deviceSynchronize()
    stop = time.time()
    print('GPU took', stop - start)
    gpu = stop - start
    print('speedup', cpu / gpu)
    print('max error v_g', cp.max(cp.abs(v_sg - cp.asarray(cpuv_sg)).ravel()))
    assert cp.allclose(v_sg, cpuv_sg, atol=1e-7, rtol=1e-13)
    print('max error dedsigmax_g',
          cp.max(cp.abs(dedsigma_xg - cp.asarray(cpudedsigma_xg)).ravel()))
    assert cp.allclose(dedsigma_xg, cpudedsigma_xg, atol=1e-7, rtol=1e-13)
    print('max error e_g', cp.max(cp.abs(e_g - cp.asarray(cpue_g)).ravel()))
    assert cp.allclose(e_g, cpue_g, atol=1e-12, rtol=1e-12)


@pytest.mark.gpu
@pytest.mark.parametrize('nspins', [1, 2])
def test_gpu_lda(nspins, gpu):
    import cupy as cp
    from gpaw.cgpaw import evaluate_lda_gpu
    ng = 10000
    n_sg = cp.exp(cp.log(10) * 5 * (cp.random.rand(nspins, ng) - 0.5))
    cp.cuda.runtime.deviceSynchronize()
    v_sg = cp.zeros_like(n_sg)
    e_g = cp.zeros((ng,))

    cpun_sg = cp.asnumpy(n_sg)
    cpuv_sg = cp.asnumpy(v_sg)
    cpue_g = cp.asnumpy(e_g)
    xc = XC('LDA')
    import time
    start = time.time()
    xc.calculate_impl(None, cpun_sg, cpuv_sg, cpue_g)
    stop = time.time()
    cpu = stop - start
    print('CPU took', stop - start)

    cp.cuda.runtime.deviceSynchronize()
    start = time.time()
    evaluate_lda_gpu(n_sg, v_sg, e_g)
    cp.cuda.runtime.deviceSynchronize()
    stop = time.time()
    print('GPU took', stop - start)
    gpu = stop - start
    print('speedup', cpu / gpu)

    assert cp.allclose(v_sg, cpuv_sg, atol=1e-13, rtol=1e-13)
    assert cp.allclose(e_g, cpue_g, atol=1e-14, rtol=1e-14)



from types import SimpleNamespace
from math import pi

import numpy as np

from gpaw.new.builder import DFTComponentsBuilder
from gpaw.new.calculation import DFTState
from gpaw.new.pwfd.ibzwfs import PWFDIBZWaveFunction
from gpaw.new.lcao.eigensolver import LCAOEigensolver
from gpaw.new.lcao.hamiltonian import LCAOHamiltonian
from gpaw.new.pwfd.davidson import Davidson
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions


class PWFDDFTComponentsBuilder(DFTComponentsBuilder):
    def __init__(self, atoms, params, *, comm, qspiral=None):
        super().__init__(atoms, params, comm=comm)
        self.qspiral_v = (None if qspiral is None else
                          qspiral @ self.grid.icell * (2 * pi))

    def create_eigensolver(self, hamiltonian):
        eigsolv_params = self.params.eigensolver.copy()
        name = eigsolv_params.pop('name', 'dav')
        assert name == 'dav'
        return Davidson(
            self.nbands,
            self.wf_desc,
            self.communicators['b'],
            hamiltonian.create_preconditioner,
            converge_bands=self.params.convergence.get('bands', 'occupied'),
            **eigsolv_params)

    def read_ibz_wave_functions(self, reader):
        kpt_comm, band_comm, domain_comm = (self.communicators[x]
                                            for x in 'kbd')

        def create_wfs(spin: int, q: int, k: int, kpt_c, weight: float):
            psit_nG = SimpleNamespace(
                comm=band_comm,
                dims=(self.nbands,),
                desc=self.wf_desc.new(kpt=kpt_c),
                data=None,
                xp=np)
            wfs = PWFDWaveFunctions(
                spin=spin,
                q=q,
                k=k,
                weight=weight,
                psit_nX=psit_nG,  # type: ignore
                setups=self.setups,
                fracpos_ac=self.fracpos_ac,
                atomdist=self.atomdist,
                ncomponents=self.ncomponents,
                qspiral_v=self.qspiral_v)

            return wfs

        ibzwfs = PWFDIBZWaveFunction.create(
            ibz=self.ibz,
            nelectrons=self.nelectrons,
            ncomponents=self.ncomponents,
            create_wfs_func=create_wfs,
            kpt_comm=self.communicators['k'],
            kpt_band_comm=self.communicators['D'],
            comm=self.communicators['w'])

        # Set eigenvalues, occupations, etc..
        self.read_wavefunction_values(reader, ibzwfs)

        return ibzwfs

    def create_ibz_wave_functions(self, basis, potential, *, log):
        from gpaw.new.lcao.builder import create_lcao_ibzwfs

        if self.params.random:
            return self.create_random_ibz_wave_functions(log)

        # sl_default = self.params.parallel['sl_default']
        # sl_lcao = self.params.parallel['sl_lcao'] or sl_default

        lcaonbands = min(self.nbands,
                         basis.Mmax * (2 if self.ncomponents == 4 else 1))
        lcao_ibzwfs, _ = create_lcao_ibzwfs(
            basis,
            self.ibz, self.communicators, self.setups,
            self.fracpos_ac, self.grid, self.dtype,
            lcaonbands, self.ncomponents, self.atomdist, self.nelectrons)

        state = DFTState(lcao_ibzwfs, None, potential)
        hamiltonian = LCAOHamiltonian(basis)
        LCAOEigensolver(basis).iterate(state, hamiltonian)

        def create_wfs(spin, q, k, kpt_c, weight):
            lcaowfs = lcao_ibzwfs.wfs_qs[q][spin]
            assert lcaowfs.spin == spin

            # Convert to PW-coefs in PW-mode:
            psit_nX = self.convert_wave_functions_from_uniform_grid(
                lcaowfs.C_nM, basis, kpt_c, q)

            mylcaonbands, nao = lcaowfs.C_nM.dist.shape
            mynbands = len(psit_nX.data)
            eig_n = np.empty(self.nbands)
            eig_n[:lcaonbands] = lcaowfs._eig_n
            eig_n[lcaonbands:] = 1e10
            if mylcaonbands < mynbands:
                psit_nX[mylcaonbands:].randomize(
                    seed=self.communicators['w'].rank)

            wfs = PWFDWaveFunctions(
                psit_nX=psit_nX,
                spin=spin,
                q=q,
                k=k,
                weight=weight,
                setups=self.setups,
                fracpos_ac=self.fracpos_ac,
                atomdist=self.atomdist,
                ncomponents=self.ncomponents,
                qspiral_v=self.qspiral_v)
            wfs._eig_n = eig_n
            return wfs

        return PWFDIBZWaveFunction.create(
            ibz=self.ibz,
            nelectrons=self.nelectrons,
            ncomponents=self.ncomponents,
            create_wfs_func=create_wfs,
            kpt_comm=self.communicators['k'],
            kpt_band_comm=self.communicators['D'],
            comm=self.communicators['w'])

    def create_random_ibz_wave_functions(self, log):
        log('Initializing wave functions with random numbers')

        def create_wfs(spin, q, k, kpt_c, weight):
            desc = self.wf_desc.new(kpt=kpt_c)
            psit_nX = desc.empty(
                dims=(self.nbands,),
                comm=self.communicators['b'],
                xp=self.xp)
            psit_nX.randomize()

            wfs = PWFDWaveFunctions(
                psit_nX=psit_nX,
                spin=spin,
                q=q,
                k=k,
                weight=weight,
                setups=self.setups,
                fracpos_ac=self.fracpos_ac,
                atomdist=self.atomdist,
                ncomponents=self.ncomponents,
                qspiral_v=self.qspiral_v)

            eig_n = self.xp.empty(self.nbands)
            eig_n[:] = 1e10  # np.inf
            wfs._eig_n = eig_n
            return wfs

        return PWFDIBZWaveFunction.create(
            ibz=self.ibz,
            nelectrons=self.nelectrons,
            ncomponents=self.ncomponents,
            create_wfs_func=create_wfs,
            kpt_comm=self.communicators['k'],
            kpt_band_comm=self.communicators['D'],
            comm=self.communicators['w'])



import numpy as np
from functools import cached_property


class ResponseKPointGrid:
    def __init__(self, kd, icell_cv, bzk_kc=None):
        # Very hacky that bzk_kc may or may not be the same as those on kd.
        # Sometimes the code likes to "just have an array", maybe to avoid
        # depending on whatever arbitrary processing is hardcoded on kd.

        self.icell_cv = icell_cv
        if bzk_kc is None:
            bzk_kc = kd.bzk_kc
        self.bzk_kc = bzk_kc
        self.bzk_kv = bzk_kc @ (2 * np.pi * icell_cv)

        # XXX May or may not be consistent with bzk_kc!
        self.kd = kd

    @cached_property
    def kptfinder(self):
        from gpaw.response.symmetry import KPointFinder
        return KPointFinder(self.bzk_kc)



import pytest
from ase import Atom, Atoms
from gpaw import GPAW


@pytest.mark.mgga
def test_mgga_mgga_restart(in_tmp_dir):
    fname = 'H2_PBE.gpw'
    fwfname = 'H2_wf_PBE.gpw'
    txt = None
    txt = '-'

    s = Atoms([Atom('H'), Atom('H', [0, 0, 1])])
    s.center(vacuum=3.)
    s.calc = GPAW(xc={'name': 'PBE', 'stencil': 1},
                  mode='fd',
                  h=.3,
                  convergence={'density': 1e-4, 'eigenstates': 1e-6})
    s.get_potential_energy()
    s.calc.write(fname)
    s.calc.write(fwfname, 'all')

    # full information
    calc = GPAW(fwfname, txt=txt)
    E_PBE = calc.get_potential_energy(s)
    dE = calc.get_xc_difference({'name': 'TPSS', 'stencil': 1})
    E_1 = E_PBE + dE
    print('E PBE, TPSS=', E_PBE, E_1)

    # no wfs
    calc = GPAW(fname, txt=txt)
    E_PBE_no_wfs = calc.get_potential_energy(s)
    dE = calc.get_xc_difference({'name': 'TPSS', 'stencil': 1})
    E_2 = E_PBE_no_wfs + dE
    print('E PBE, TPSS=', E_PBE_no_wfs, E_2)

    print('diff=', E_1 - E_2)
    assert abs(E_1 - E_2) < 0.005

    energy_tolerance = 0.002
    assert E_PBE == pytest.approx(-5.341, abs=energy_tolerance)
    assert E_PBE_no_wfs == pytest.approx(-5.33901, abs=energy_tolerance)
    assert E_1 == pytest.approx(-5.57685, abs=energy_tolerance)
    assert E_2 == pytest.approx(-5.57685, abs=energy_tolerance)



import numpy as np
from gpaw import GPAW, restart
import pytest


def get_restart_test_values(calc):
    # XXX the forces must be evaluated first because a force evaluation
    # somehow affects (changes) both the energy and the wf when restarting
    # from a gpw file. An issue was created (#1051) documenting this unexpected
    # and rather bizarre behavior.

    atoms = calc.get_atoms()
    f = atoms.get_forces()
    e = atoms.get_potential_energy()
    m = atoms.get_magnetic_moments()

    eig0 = calc.get_eigenvalues(spin=0)
    eig1 = calc.get_eigenvalues(spin=1)

    return e, f, m, eig0, eig1


def test_fileio_restart_density(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['na3_fd_density_restart'])

    e0, f0, m0, eig00, eig01 = get_restart_test_values(calc=calc)

    # Write the restart file
    calc.write('tmp.gpw')

    # Try restarting from all the files
    atoms, calc = restart('tmp.gpw')
    e1, f1, m1, eig10, eig11 = get_restart_test_values(calc=calc)

    print(e0, e1)
    assert e0 == pytest.approx(e1, abs=2e-3)
    print(f0, f1)
    for ff0, ff1 in zip(f0, f1):
        err = np.linalg.norm(ff0 - ff1)
        # for forces, we use larger tolerance
        assert err == pytest.approx(0.0, abs=4e-2)
    print(m0, m1)
    for mm0, mm1 in zip(m0, m1):
        assert mm0 == pytest.approx(mm1, abs=2e-3)
    print('A', eig00, eig10)
    for eig0, eig1 in zip(eig00, eig10):
        assert eig0 == pytest.approx(eig1, abs=5e-3)
    print('B', eig01, eig11)
    for eig0, eig1 in zip(eig01, eig11):
        assert eig0 == pytest.approx(eig1, abs=2e-2)

    # Check that after restart, everything is writable
    calc.write('tmp2.gpw')
